#include "Main.h"

#ifdef AVI_TEST_MAIN
#include "TestMain.h"
#endif

#include <MainLoop.hpp>
#include <OS.hpp>
#include <arviz_godot/lib/avi/app/ArVizInteractive.h>
#include <arviz_godot/lib/avi/core/Settings.h>
#include <arviz_godot/lib/avi/gui/UserInterface.h>
#include <arviz_godot/lib/avi/scene/OrientationIndicator.h>
#include <arviz_godot/lib/avi/scene/VisualizedScene.h>

void
avi::app::Main::_register_methods()
{
    register_method("_ready", &Main::_ready);
    register_method("_process", &Main::_process);
    register_method("_notification", &Main::_notification);
}

avi::app::Main* avi::app::Main::instance = nullptr;

void
avi::app::Main::_init()
{
    instance = this;

    bool mockStorage = false;
    godot::OS* os = godot::OS::get_singleton();

    godot::PoolStringArray args = os->get_cmdline_args();
    for (int i = 0; i < args.size(); i++)
    {
        if (args[i] == "--mock-storage")
        {
            mockStorage = true;
        }
    }

    avi::core::Settings settings = avi::core::Settings::getDefaultSettings();
    appSettings = std::make_unique<avi::app::ApplicationSettings>(std::move(settings));
    appSettings->getSettings().save();

    eventList = std::make_unique<avi::app::EventList>();

    avi::connection::Component* component;

    if (!mockStorage)
    {
        this->component = avi::app::ArVizInteractive::Initialize(*appSettings);
        component = this->component;

        ARMARX_INFO << "Initialized ArVizInteractive" << std::endl;
    }
    else
    {
        component = new avi::connection::MockStorage::MockComponent();
    }

    revisions = std::make_unique<avi::connection::RevisionBuffer>();
    revisionLoadCycle =
        std::make_unique<avi::app::RevisionLoadCycle>(*component, *revisions, *appSettings);

    revisionLoadCycle->run(mockStorage);
}

void
avi::app::Main::_ready()
{
    godot::OS::get_singleton()->set_min_window_size(minimumWindowSize);

    keyInputEventDispatch = get_node<avi::app::KeyInputEventDispatch>("Key Input Events");
    keyInputEventDispatch->hookup(*eventList);

    graphicsController = std::make_unique<avi::scene::GraphicsController>(this);
    profileStorage = std::make_unique<avi::scene::ProfileStorage>(*appSettings);

    replayController =
        std::make_unique<avi::replay::ReplayController>(revisionLoadCycle->getStorage());

    auto* scene = get_node<avi::scene::VisualizedScene>("3D/3D/Visualized Scene");
    scene->setup(*revisions, graphicsController.get());

    auto defaultProfile = profileStorage->retrieveDefault();
    if (defaultProfile)
        graphicsController->activateProfile(*defaultProfile);
    else
        graphicsController->activateProfile(profileStorage->createDefault());

    auto* userInterface = godot::Object::cast_to<avi::gui::UserInterface>(get_node("GUI"));

    viewCamera = get_node<avi::scene::ViewCamera>("3D/3D/View Camera");
    viewCamera->setup(*scene, *eventList, *appSettings, *graphicsController);

    get_node<avi::scene::OrientationIndicator>("OrientationIndicator")->setup(scene, viewCamera);

    cameraSet = std::make_unique<avi::scene::CameraSet>(viewCamera, *eventList);

    imageRecorder = std::make_unique<avi::recording::ImageRecorder>(
        *cameraSet, *appSettings, *eventList); // used through events
    videoRecorder =
        std::make_unique<avi::recording::VideoRecorder>(*cameraSet, *appSettings, *eventList);

    interactionDispatch = get_node<avi::interaction::InteractionDispatch>("Interaction Dispatch");
    interactionDispatch->setup(revisionLoadCycle->getStorage(),
                               *cameraSet,
                               *userInterface,
                               revisionLoadCycle->getStorage());

    userInterface->setupUI(*eventList, *eventList, appSettings->getSettings());
    userInterface->setupInteraction(*interactionDispatch);
    userInterface->setupManipulation();
    userInterface->setupStatusIcons(
        revisionLoadCycle->getStorage(), revisionLoadCycle->getStorage(), *videoRecorder);
    userInterface->setupScene(*scene);
    userInterface->setupGraphics(*graphicsController, *profileStorage);
    userInterface->setupReplay(*replayController, *scene);

#ifdef AVI_TEST_MAIN
    testMain();
#endif
}

void
avi::app::Main::_process()
{
    videoRecorder->run();
}

void
avi::app::Main::_notification(int what)
{
    if (what == godot::MainLoop::NOTIFICATION_WM_QUIT_REQUEST)
    {
        ARMARX_INFO << "Quitting...";

        revisionLoadCycle->stop();

        if (component != nullptr)
        {
            component->close();
        }

        profileStorage->put(graphicsController->getActiveProfile());
        profileStorage->cleanup();

        ARMARX_INFO << "All cleanup on exit done";
    }

    if (what == godot::MainLoop::NOTIFICATION_OS_MEMORY_WARNING)
    {
        ARMARX_WARNING << "Godot reported memory warning!";
    }

    if (what == godot::MainLoop::NOTIFICATION_CRASH)
    {
        ARMARX_FATAL << "Godot crashed!";
    }
}

avi::app::Main*
avi::app::Main::get()
{
    return instance;
}
