#include "ReplayStorageAdapter.h"

#include <utility>

avi::connection::ReplayStorageAdapter::ReplayStorageAdapter(
    avi::connection::Component& component,
    std::shared_ptr<avi::connection::Clock> clock,
    std::string replay) :
    component(component), clock(std::move(clock)), activeReplay(std::move(replay))
{
}

armarx::viz::data::LayerUpdates
avi::connection::ReplayStorageAdapter::getLayerUpdates()
{
    try
    {
        auto time = std::chrono::duration_cast<std::chrono::microseconds>(
                        std::chrono::steady_clock::now() - last)
                        .count();
        clock->advance(time);
        armarx::viz::data::LayerUpdates updates;
        int64_t wantedTimeStamp = clock->getTargetTime() + clock->getTimeFrame().begin;

        if (wantedTimeStamp > clock->getTimeFrame().end ||
            wantedTimeStamp < clock->getTimeFrame().begin)
        {
            return updates;
        }

        int64_t index = -1;
        for (auto& [key, value] : batches)
        {
            if (key > wantedTimeStamp)
            {
                // get index of batch containing wantedTimeStamp
                index = value - 1;
                break;
            }
        }
        if (index == -1)
        {
            index = static_cast<int64_t>(batches.size() - 1);
        }

        // calculate wanted batches
        std::vector<int64_t> wanted;
        int signum = 1;
        if (clock->getSpeed() < 0)
        {
            signum = -1;
        }

        for (int64_t i = -(signum * BATCHES_BACK); i != (signum * (BATCHES_FRONT + 1));
             i = i + signum)
        {
            if ((index + i >= 0) && (index + i < static_cast<long>(batches.size())))
            {
                wanted.push_back(index + i);
            }
        }

        // delete all batches no longer needed
        for (auto itCache = batchCache.begin(); itCache != batchCache.end();)
        {
            auto& [key, value] = *itCache;
            bool found = false;
            for (auto it = wanted.begin(); it != wanted.end();)
            {
                if (key == *it)
                {
                    found = true;
                    it = wanted.erase(it);
                }
                else
                {
                    it++;
                }
            }
            if (!found)
            {
                itCache = batchCache.erase(itCache);
            }
            else
            {
                itCache++;
            }
        }

        // load all missing batches
        for (int64_t batchIndex : wanted)
        {
            batchCache.insert(std::make_pair(
                batchIndex,
                component.getStorageInterface()->getRecordingBatch(activeReplay, batchIndex)));
        }

        armarx::viz::data::RecordingBatch currentBatch = batchCache.at(index);
        size_t i =
            binarySearch(currentBatch.updates, 0, currentBatch.updates.size(), wantedTimeStamp);
        wantedTimeStamp = currentBatch.updates[i].timestampInMicroseconds;
        updates.updates.push_back(currentBatch.updates[i].update);
        while (i > 0 && currentBatch.updates[i - 1].timestampInMicroseconds == wantedTimeStamp)
        {
            i--;
            updates.updates.push_back(currentBatch.updates[i].update);
        }

        last = std::chrono::steady_clock::now();
        return updates;
    }
    catch (IceUtil::Exception& e)
    {
        ARMARX_ERROR << "Error while pulling updates from storage: " << e.what();
        return {};
    }
}

size_t
avi::connection::ReplayStorageAdapter::binarySearch(
    const std::vector<armarx::viz::data::TimestampedLayerUpdate>& updates,
    size_t l,
    size_t r,
    int64_t timeStamp)
{
    if (r >= l)
    {
        size_t mid = l + (r - l) / 2;
        if ((updates[mid].timestampInMicroseconds <= timeStamp) &&
            (mid + 1 == updates.size() || updates[mid + 1].timestampInMicroseconds > timeStamp))
        {
            return mid;
        }
        else if (updates[mid].timestampInMicroseconds > timeStamp)
        {
            return binarySearch(updates, l, mid - 1, timeStamp);
        }
        else
        {
            return binarySearch(updates, mid + 1, r, timeStamp);
        }
    }
    else
    {
        return -1;
    }
}

std::vector<std::string>
avi::connection::ReplayStorageAdapter::getAllReplays()
{
    try
    {
        std::vector<armarx::viz::data::Recording> replays =
            component.getStorageInterface()->getAllRecordings();

        std::vector<std::string> replayNames;
        replayNames.reserve(replays.size());

        for (const auto& replay : replays)
        {
            replayNames.push_back(replay.id);
        }

        return replayNames;
    }
    catch (IceUtil::Exception& e)
    {
        ARMARX_ERROR << "Error while pulling replay names: " << e.what();
        return {};
    }
}

void
avi::connection::ReplayStorageAdapter::setup()
{
    try
    {
        auto recordings = component.getStorageInterface()->getAllRecordings();

        if (recordings.empty())
            return;

        size_t index = 0;
        while (recordings[index].id != activeReplay)
        {
            index++;
        }
        clock->setTimeFrame(TimeFrame(recordings[index].firstTimestampInMicroSeconds,
                                      recordings[index].lastTimestampInMicroSeconds));
        for (size_t i = 0; i < recordings[index].batchHeaders.size(); i++)
        {
            batches.insert(
                std::make_pair(recordings[index].batchHeaders[i].firstTimestampInMicroSeconds, i));
        }
        armarx::viz::data::RecordingBatch batch =
            component.getStorageInterface()->getRecordingBatch(activeReplay, 0);

        batchCache.insert(
            std::make_pair<size_t, armarx::viz::data::RecordingBatch>(0, std::move(batch)));
        clock->setAvailableTimeFrame(
            TimeFrame(recordings[index].batchHeaders[0].firstTimestampInMicroSeconds,
                      recordings[index].batchHeaders[0].lastTimestampInMicroSeconds));
    }

    catch (IceUtil::Exception& e)
    {
        ARMARX_ERROR << "Error while loading replay information: " << e.what();
    }
}

void
avi::connection::ReplayStorageAdapter::setInteraction(
    std::vector<avi::connection::Interaction> newInteractions)
{
}
