#include "GraphicsSettings.h"

#include <Button.hpp>
#include <Godot.hpp>
#include <OptionButton.hpp>

void
avi::gui::GraphicsSettings::_register_methods()
{
    register_method("_ready", &GraphicsSettings::_ready);

    register_method("_save", &GraphicsSettings::_save);
    register_method("_copy", &GraphicsSettings::_copy);

    register_method("_select", &GraphicsSettings::_select);

    register_method("_change", &GraphicsSettings::_change);
}

void
avi::gui::GraphicsSettings::_init()
{
}

void
avi::gui::GraphicsSettings::_ready()
{
    saveButton = get_node<godot::Button>("Background/Content/Buttons/Save");
    saveButton->connect("pressed", this, "_save");

    get_node<godot::Button>("Background/Content/Buttons/Copy")->connect("pressed", this, "_copy");

    options = get_node<godot::OptionButton>("Background/Content/Profiles");
    options->connect("item_selected", this, "_select");

    nameBox = get_node<godot::LineEdit>("Background/Content/Name");

    container = get_node<godot::Control>("Background/Content/Scroll/Container");
}

void
avi::gui::GraphicsSettings::setup(avi::scene::GraphicsController& controller,
                                  avi::scene::ProfileStorage& storage)
{
    this->controller = &controller;
    this->storage = &storage;

    update();
    setSavedFlag(true);
}

void
avi::gui::GraphicsSettings::_save()
{
    std::string newName = getName();

    if (newName.empty() || storage->has(newName))
        return;

    auto& profile = this->controller->getActiveProfile();

    profile.rename(newName);
    storage->save(profile);

    update();
    setSavedFlag(true);
}

void
avi::gui::GraphicsSettings::_copy()
{
    avi::scene::Profile original = this->controller->getActiveProfile();
    avi::scene::Profile copy = original;

    copy.rename("Copy of " + original.getName());

    this->controller->activateProfile(copy);
    copy = this->controller->getActiveProfile(); // Controller might have corrected some values

    storage->put(original);
    storage->save(copy);

    update();
    setSavedFlag(true);
}

void
avi::gui::GraphicsSettings::_select(int64_t idx)
{
    std::string selectedName = options->get_item_text(idx).ascii().get_data();

    auto profile = this->storage->retrieve(selectedName);

    if (profile)
    {
        auto previous = this->controller->getActiveProfile();
        storage->put(previous);

        this->controller->activateProfile(*profile);
    }

    update();
    setSavedFlag(true);
}

void
avi::gui::GraphicsSettings::_change(godot::Variant v)
{
    for (auto& setter : setters)
    {
        setter();
    }

    this->controller->update();
    setSavedFlag(false);
}

void
avi::gui::GraphicsSettings::update()
{
    options->clear();

    std::string activeName = this->controller->getActiveProfile().getName();
    options->add_item(activeName.c_str());
    setName(activeName);

    for (auto& name : storage->getNames())
    {
        options->add_item(name.c_str());
    }

    for (int64_t child = 0; child < container->get_child_count(); child++)
    {
        container->get_child(child)->queue_free();
    }

    VisualOptionTransformer transformer;
    optionStore = this->controller->getActiveProfile().getOptions<VisualOptionTransformer>();

    std::vector<avi::core::Option<VisualOptionTransformer>*> optionsToTransform;
    optionsToTransform.reserve(optionStore.size());
    for (auto& option : optionStore)
    {
        optionsToTransform.push_back(option.get());
    }

    setters.clear();
    setters = transformer.createGraphicsControls(container, this, std::move(optionsToTransform));
}

void
avi::gui::GraphicsSettings::setName(const std::string& name)
{
    nameBox->set_text(name.c_str());
}

std::string
avi::gui::GraphicsSettings::getName()
{
    godot::String name = nameBox->get_text();
    return name.ascii().get_data();
}

void
avi::gui::GraphicsSettings::setSavedFlag(bool flag)
{
    saveButton->set_text(flag ? "Save" : "Save (*)");
}
