#include "LayerSelection.h"

#include <Label.hpp>
#include <TreeItem.hpp>


void
avi::gui::LayerSelection::_register_methods()
{
    register_method("_ready", &LayerSelection::_ready);
    register_method("update_layers", &LayerSelection::update_layers);
    register_method("setVisibility", &LayerSelection::setVisibility);
    register_method("updateActiveComponents", &LayerSelection::updateActiveComponents);
    register_method("setVisibilityAll", &LayerSelection::setVisibilityAll);
    register_method("setCollapseAll", &LayerSelection::setCollapseAll);
    register_method("filterLayers", &LayerSelection::filterLayers);
    register_method("setFilterVisibility", &LayerSelection::setFilterVisibility);
}

void
avi::gui::LayerSelection::_init()
{
}

void
avi::gui::LayerSelection::_ready()
{
    container = get_node<godot::Control>("Background/Content/Scroll/Layers");
    hideAllButton = get_node<godot::Control>("Background/Content/Buttons/HideAllButton");
    showAllButton = get_node<godot::Control>("Background/Content/Buttons/ShowAllButton");

    collapseAllButton = get_node<godot::Control>("Background/Content/Buttons/CollapseAllButton");
    expandAllButton = get_node<godot::Control>("Background/Content/Buttons/ExpandAllButton");

    filterTextBox = get_node<godot::LineEdit>("Background/Content/FilterButtons/FilterTextBox");
    hideFilteredButton =
        get_node<godot::Control>("Background/Content/FilterButtons/HideFilteredButton");
    showFilteredButton =
        get_node<godot::Control>("Background/Content/FilterButtons/ShowFilteredButton");
}

void
avi::gui::LayerSelection::setup(avi::scene::VisualizedScene& scene)
{
    this->scene = &scene;

    scene.connect("layers_changed", this, "update_layers");

    godot::Array args = godot::Array::make(false);
    hideAllButton->connect("button_down", this, "setVisibilityAll", args);
    expandAllButton->connect("button_down", this, "setCollapseAll", args);
    hideFilteredButton->connect("button_down", this, "setFilterVisibility", args);
    args[0] = true;
    showAllButton->connect("button_down", this, "setVisibilityAll", args);
    collapseAllButton->connect("button_down", this, "setCollapseAll", args);
    showFilteredButton->connect("button_down", this, "setFilterVisibility", args);

    filterTextBox->connect("text_changed", this, "filterLayers");
}

void
avi::gui::LayerSelection::update_layers()
{
    std::lock_guard<std::mutex> lock(mutex);

    for (int64_t child = 0; child < container->get_child_count(); child++)
    {
        container->get_child(child)->queue_free();
    }

    components = scene->getComponents();
    filterLayers();
}


void
avi::gui::LayerSelection::updateActiveComponents()
{
    for (auto& layerTreeItem : layerTreeItems)
    {
        scene->setLayerVisibility(components[layerTreeItem.componentIndex],
                                  layers[layerTreeItem.componentIndex][layerTreeItem.layerIndex],
                                  layerTreeItem.treeItem->is_checked(0));
    }

    int component = 0;
    for (auto& componentTreeItem : componentTreeItems)
    {
        if ((!componentTreeItem.treeItem->is_checked(0) && componentTreeItem.previousState) ||
            (componentTreeItem.treeItem->is_checked(0) && !componentTreeItem.previousState))
        {

            for (auto& layerTreeItem : layerTreeItems)
            {
                if (layerTreeItem.componentIndex != component)
                {
                    continue;
                }

                if (layerTreeItem.treeItem->get_parent() == componentTreeItem.treeItem)
                {
                    layerTreeItem.treeItem->set_checked(0,
                                                        componentTreeItem.treeItem->is_checked(0));
                }

                if (layerTreeItem.active)
                {
                    scene->setLayerVisibility(components[component],
                                              layers[component][layerTreeItem.layerIndex],
                                              componentTreeItem.treeItem->is_checked(0));
                }
            }

            scene->setComponentVisibility(components[component],
                                          componentTreeItem.treeItem->is_checked(0));
        }

        componentTreeItem.previousState = componentTreeItem.treeItem->is_checked(0);
        component++;
    }
}


void
avi::gui::LayerSelection::setVisibility(bool visible, int64_t component, int64_t layer)
{
    scene->setLayerVisibility(components[component], layers[component][layer], visible);
}

void
avi::gui::LayerSelection::setVisibilityAll(bool visible)
{
    for (auto& componentTreeItem : componentTreeItems)
    {
        componentTreeItem.treeItem->set_checked(0, visible);
    }

    for (auto& layerTreeItem : layerTreeItems)
    {
        layerTreeItem.treeItem->set_checked(0, visible);
    }

    updateActiveComponents();
}

void
avi::gui::LayerSelection::setCollapseAll(bool collapsed)
{

    for (auto& componentTreeItem : componentTreeItems)
    {
        componentTreeItem.treeItem->set_collapsed(collapsed);
    }

    for (auto& layerTreeItem : layerTreeItems)
    {
        layerTreeItem.treeItem->set_collapsed(collapsed);
    }
}

void
avi::gui::LayerSelection::setFilterVisibility(bool visible)
{
    filterLayers();

    for (auto& layerTreeItem : layerTreeItems)
    {
        if (layerTreeItem.active)
        {
            layerTreeItem.treeItem->set_checked(0, visible);
        }
    }

    updateActiveComponents();
}

void
avi::gui::LayerSelection::filterLayers()
{
    godot::String filterName = filterTextBox->get_text();

    buildTree();

    if (filterName.to_lower() == "")
    {
        return;
    }

    int componentIndex = 0;
    for (auto& componentTreeItem : componentTreeItems)
    {
        for (auto& layerTreeItem : layerTreeItems)
        {
            if (layerTreeItem.componentIndex != componentIndex)
            {
                continue;
            }

            if (layerTreeItem.treeItem->get_text(0).to_lower().find(filterName.to_lower()) == -1)
            {
                componentTreeItem.treeItem->remove_child(layerTreeItem.treeItem);
                layerTreeItem.active = false;
            }
        }

        if (componentTreeItem.treeItem->get_children() == nullptr)
        {
            treeRootNode->remove_child(componentTreeItem.treeItem);
        }

        componentIndex++;
    }
}

void
avi::gui::LayerSelection::buildTree()
{
    if (tree != nullptr)
    {
        tree->queue_free();
    }

    tree = godot::Tree::_new();
    container->add_child(tree);
    godot::Vector2 tree_min_size = godot::Vector2(339, 430);
    tree->set_custom_minimum_size(tree_min_size);
    tree->set_h_size_flags(3);
    tree->set_v_size_flags(3);
    treeRootNode = tree->create_item();
    tree->set_hide_root(true);

    for (auto& layerTreeItem : layerTreeItems)
    {
        layerTreeItem.treeItem->free();
    }

    for (auto& componentTreeItem : componentTreeItems)
    {
        componentTreeItem.treeItem->free();
    }

    componentTreeItems.clear();
    layerTreeItems.clear();
    layers.clear();

    int componentIndex = 0;
    for (auto& component : components)
    {
        bool componentVisibility = scene->getComponentVisibility(components[componentIndex]);

        godot::TreeItem* componentTreeItem = tree->create_item(treeRootNode);
        componentTreeItem->set_cell_mode(0, godot::TreeItem::CELL_MODE_CHECK);
        componentTreeItem->set_checked(0, componentVisibility);
        componentTreeItem->set_editable(0, true);
        componentTreeItem->set_tooltip(0, "Show / Hide a Layer");
        componentTreeItem->set_text(0, component.c_str());
        componentTreeItems.push_back({componentTreeItem, componentVisibility});

        layers.push_back(scene->getLayers(component));

        int layerIndex = 0;
        for (auto& layer : layers.back())
        {
            godot::TreeItem* layerTreeItem = tree->create_item(componentTreeItem);
            layerTreeItem->set_cell_mode(0, 1); //mode 1 = checkbox
            layerTreeItem->set_checked(
                0,
                scene->getLayerVisibility(components[componentIndex],
                                          layers[componentIndex][layerIndex]));
            layerTreeItem->set_editable(0, true);
            layerTreeItem->set_tooltip(0, "Show / Hide a Layer");
            layerTreeItem->set_text(0, layer.c_str());
            layerTreeItems.push_back({layerTreeItem, componentIndex, layerIndex, true});
            layerIndex++;
        }
        componentIndex++;
    }
    tree->connect("item_edited", this, "updateActiveComponents");
}

void
avi::gui::LayerSelection::clear()
{
    layers.clear();
    components.clear();
}

void
avi::gui::LayerSelection::lock()
{
    mutex.lock();
}

void
avi::gui::LayerSelection::unlock()
{
    mutex.unlock();
}
