#include "UserInterface.h"

#include <Input.hpp>
#include <InputEvent.hpp>
#include <MenuButton.hpp>
#include <PopupMenu.hpp>
#include <arviz_godot/lib/avi/gui/CameraModeMenu.h>
#include <arviz_godot/lib/avi/gui/SelectionMenu.h>
#include <arviz_godot/lib/avi/gui/StatusIcons.h>

void
avi::gui::UserInterface::_register_methods()
{
    register_method("_ready", &UserInterface::_ready);
    register_method("_unhandled_input", &UserInterface::_unhandled_input);
    register_method("_scene_menu_select", &UserInterface::_scene_menu_select);
    register_method("_view_menu_select", &UserInterface::_view_menu_select);
}

void
avi::gui::UserInterface::_init()
{
}

void
avi::gui::UserInterface::_ready()
{
    layerSelection = get_node<avi::gui::LayerSelection>("Overlays/Layers");
    graphicsSettings = get_node<avi::gui::GraphicsSettings>("Overlays/Visuals");
    replayCockpit = get_node<avi::gui::ReplayCockpit>("Overlays/Replay");

    settingsWindow = get_node<avi::gui::SettingsWindow>("Settings");
    orientationWindow = get_node<avi::gui::OrientationWindow>("Orientation");

    get_node<godot::MenuButton>("Bar/Bar Background/Bar Container/Scene Menu")
        ->get_popup()
        ->connect("id_pressed", this, "_scene_menu_select");
    get_node<godot::MenuButton>("Bar/Bar Background/Bar Container/View Menu")
        ->get_popup()
        ->connect("id_pressed", this, "_view_menu_select");

    helpWindow = get_node<godot::Popup>("Help");
}

void
avi::gui::UserInterface::_unhandled_input(godot::InputEvent* event)
{
    if (event->is_action_pressed("avi_hide_ui"))
    {
        if (hiddenToggle)
            requestVisibleUI();
        else
            requestHiddenUI();

        hiddenToggle = !hiddenToggle;
    }
}

void
avi::gui::UserInterface::setupUI(avi::core::EventInvokingSystem& eventInvokingSystem,
                                 avi::core::EventSubscriptionSystem& eventSubscriptionSystem,
                                 avi::core::Settings& settings)
{
    this->eventInvokingSystem = &eventInvokingSystem;
    this->eventSubscriptionSystem = &eventSubscriptionSystem;

    settingsWindow->setup(settings);

    get_node<avi::gui::CameraModeMenu>("Bar/Bar Background/Bar Container/CameraMenu")
        ->setup(eventInvokingSystem, eventSubscriptionSystem);
}

void
avi::gui::UserInterface::setupStatusIcons(avi::connection::Connectable& connectable,
                                          avi::connection::StorageModeSwitch& mode,
                                          avi::recording::VideoRecorder& recorder)
{
    this->connectable = &connectable;

    auto* statusIcons = godot::Object::cast_to<avi::gui::StatusIcons>(get_node("StatusBar"));

    statusIcons->setup(connectable, mode, recorder);
}

void
avi::gui::UserInterface::setupScene(avi::scene::VisualizedScene& visualizedScene)
{
    layerSelection->setup(visualizedScene);
    orientationWindow->setup(visualizedScene);
}

void
avi::gui::UserInterface::setupGraphics(avi::scene::GraphicsController& graphicsController,
                                       avi::scene::ProfileStorage& profileStorage)
{
    graphicsSettings->setup(graphicsController, profileStorage);
}

void
avi::gui::UserInterface::setupInteraction(
    avi::interaction::InteractionDispatch& interactionDispatch)
{
    this->interactionDispatch = &interactionDispatch;
    this->selectionMenu = get_node<avi::gui::SelectionMenu>("InteractionPopupMenu");
}

void
avi::gui::UserInterface::setupReplay(avi::replay::ReplayController& replayController,
                                     avi::scene::VisualizedScene& visualizedScene)
{
    this->replayCockpit->setup(replayController, visualizedScene, this);
}

void
avi::gui::UserInterface::setupManipulation()
{
    manipulationMenu =
        get_node<avi::gui::ManipulationMenu>("Bar/Bar Background/Bar Container/ManipulationMenu");
    manipulationMenu->setup(*interactionDispatch->getManipulator(), *eventSubscriptionSystem);
}

avi::core::EventInvokingSystem*
avi::gui::UserInterface::getEventSystem()
{
    return eventInvokingSystem;
}

void
avi::gui::UserInterface::_scene_menu_select(int64_t id)
{
    switch (id)
    {
        case 0: // Orientation
            orientationWindow->open();
            break;

        case 1: // Settings
            settingsWindow->open();
            break;

        case 2: // Help
            helpWindow->popup_centered();
            break;
    }
}

void
avi::gui::UserInterface::_view_menu_select(int64_t id)
{
    switch (id)
    {
        case 0: // Reset View
            eventInvokingSystem->invoke("VIEW_CAMERA_RESET");
            break;

        case 1: // Focus All
            eventInvokingSystem->invoke("VIEW_CAMERA_FOCUS_SCENE");
            break;

        case 2: // Take Screenshot
            eventInvokingSystem->invoke("TAKE_SCREENSHOT");
            break;

        case 3: // Record Video
            eventInvokingSystem->invoke("RECORD_VIDEO");
            break;

        case 4: // Hide UI
            requestHiddenUI();
            hiddenToggle = true;
            break;
    }
}

void
avi::gui::UserInterface::requestHiddenUI()
{
    hidingCounter++;

    if (hidingCounter > 0)
    {
        set_visible(false);
    }
}

void
avi::gui::UserInterface::requestVisibleUI()
{
    assert(hidingCounter > 0 && "More visible requests than hidden requests.");

    hidingCounter--;

    if (hidingCounter == 0)
    {
        set_visible(true);
    }
}

void
avi::gui::UserInterface::openSelectionMenu(godot::Vector2 mousePosition,
                                           std::vector<std::string> interactions,
                                           std::function<void(int64_t)> callback)
{
    this->selectionMenu->open(mousePosition, interactions, callback);
}

void
avi::gui::UserInterface::openTextInputPrompt()
{
}

void
avi::gui::UserInterface::clear()
{
    layerSelection->clear();
}

void
avi::gui::UserInterface::lockLayerSelection()
{
    layerSelection->lock();
}

void
avi::gui::UserInterface::unlockLayerSelection()
{
    layerSelection->unlock();
}
