#pragma once

#include <filesystem>
#include <functional>
#include <vector>

#include <CheckButton.hpp>
#include <ColorPickerButton.hpp>
#include <Control.hpp>
#include <Godot.hpp>
#include <Label.hpp>
#include <LineEdit.hpp>
#include <OptionButton.hpp>
#include <PackedScene.hpp>
#include <ResourceLoader.hpp>
#include <SpinBox.hpp>
#include <Viewport.hpp>
#include <arviz_godot/lib/avi/core/ValueOption.h>
#include <arviz_godot/lib/avi/scene/DrawStyle.h>

namespace avi::gui
{
    class VisualOptionTransformer
    {
    private:
        template <typename C>
        static C*
        create(godot::Control* root,
               const std::string& resource,
               const std::string& label,
               const std::string& node)
        {
            godot::ResourceLoader* loader = godot::ResourceLoader::get_singleton();
            godot::Ref<godot::PackedScene> element = loader->load(resource.c_str());

            auto* elementRoot = godot::Object::cast_to<godot::Control>(element->instance());

            elementRoot->set_name(label.c_str());
            elementRoot->get_node<godot::Label>("Label")->set_text(label.c_str());

            root->add_child(elementRoot);

            return elementRoot->get_node<C>(node.c_str());
        }

    public:
        std::vector<std::function<void()>>
        createGraphicsControls(godot::Control* root,
                               godot::Node* receiver,
                               std::vector<avi::core::Option<VisualOptionTransformer>*>&& options);

        template <typename V>
        void transform(avi::core::ValueOption<V, VisualOptionTransformer>& option);

        template <typename V, typename D = void>
        struct Transform
        {
            static void
            transform(avi::core::ValueOption<V, VisualOptionTransformer>& option,
                      godot::Control* root,
                      godot::Node* receiver,
                      std::vector<std::function<void()>>& fns)
            {
            }
        };

        template <typename D>
        struct Transform<godot::Color, D>
        {
            static void
            transform(avi::core::ValueOption<godot::Color, VisualOptionTransformer>& option,
                      godot::Control* root,
                      godot::Node* receiver,
                      std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::ColorPickerButton>(
                    root, "res://scenes/options/color.tscn", option.getLabel(), "Color");
                values->connect("color_changed", receiver, "_change");

                values->set_pick_color(option.getValue());

                std::function<void()> fn = [values, &option]()
                { option.setValue(values->get_pick_color()); };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<double, D>
        {
            static void
            transform(avi::core::ValueOption<double, VisualOptionTransformer>& option,
                      godot::Control* root,
                      godot::Node* receiver,
                      std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::SpinBox>(
                    root, "res://scenes/options/double.tscn", option.getLabel(), "Double");
                values->connect("value_changed", receiver, "_change");

                values->set_value(option.getValue());

                std::function<void()> fn = [values, &option]()
                { option.setValue(values->get_value()); };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<bool, D>
        {
            static void
            transform(avi::core::ValueOption<bool, VisualOptionTransformer>& option,
                      godot::Control* root,
                      godot::Node* receiver,
                      std::vector<std::function<void()>>& fns)
            {
                auto* value = create<godot::CheckButton>(
                    root, "res://scenes/options/boolean.tscn", option.getLabel(), "Boolean");
                value->connect("toggled", receiver, "_change");

                value->set_pressed(option.getValue());

                std::function<void()> fn = [value, &option]()
                { option.setValue(value->is_pressed()); };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<godot::Viewport::MSAA, D>
        {
            static void
            transform(
                avi::core::ValueOption<godot::Viewport::MSAA, VisualOptionTransformer>& option,
                godot::Control* root,
                godot::Node* receiver,
                std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::OptionButton>(
                    root, "res://scenes/options/option.tscn", option.getLabel(), "Option");
                values->connect("item_selected", receiver, "_change");

                values->add_item("Disabled");
                values->add_item("2x");
                values->add_item("4x");
                values->add_item("8x");
                values->add_item("16x");

                int64_t select = 0;

                switch (option.getValue())
                {
                    case godot::Viewport::MSAA_DISABLED:
                        select = 0;
                        break;
                    case godot::Viewport::MSAA_2X:
                        select = 1;
                        break;
                    case godot::Viewport::MSAA_4X:
                        select = 2;
                        break;
                    case godot::Viewport::MSAA_8X:
                        select = 3;
                        break;
                    case godot::Viewport::MSAA_16X:
                        select = 4;
                        break;
                }

                values->select(select);

                std::function<void()> fn = [values, &option]()
                {
                    auto idx = values->get_selected();
                    switch (idx)
                    {
                        case 0:
                            option.setValue(godot::Viewport::MSAA::MSAA_DISABLED);
                            break;
                        case 1:
                            option.setValue(godot::Viewport::MSAA::MSAA_2X);
                            break;
                        case 2:
                            option.setValue(godot::Viewport::MSAA::MSAA_4X);
                            break;
                        case 3:
                            option.setValue(godot::Viewport::MSAA::MSAA_8X);
                            break;
                        case 4:
                            option.setValue(godot::Viewport::MSAA::MSAA_16X);
                            break;

                        default:
                            break;
                    }
                };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<avi::scene::DrawStyle, D>
        {
            static void
            transform(
                avi::core::ValueOption<avi::scene::DrawStyle, VisualOptionTransformer>& option,
                godot::Control* root,
                godot::Node* receiver,
                std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::OptionButton>(
                    root, "res://scenes/options/option.tscn", option.getLabel(), "Option");
                values->connect("item_selected", receiver, "_change");

                values->add_item("Original");
                values->add_item("Visualization");
                values->add_item("Collision");

                int64_t select = 0;

                switch (option.getValue())
                {
                    case avi::scene::DrawStyle::ORIGINAL:
                        select = 0;
                        break;
                    case avi::scene::DrawStyle::VISUALIZATION:
                        select = 1;
                        break;
                    case avi::scene::DrawStyle::COLLISION:
                        select = 2;
                        break;
                }

                values->select(select);

                std::function<void()> fn = [values, &option]()
                {
                    auto idx = values->get_selected();
                    switch (idx)
                    {
                        case 0:
                            option.setValue(avi::scene::DrawStyle::ORIGINAL);
                            break;
                        case 1:
                            option.setValue(avi::scene::DrawStyle::VISUALIZATION);
                            break;
                        case 2:
                            option.setValue(avi::scene::DrawStyle::COLLISION);
                            break;

                        default:
                            break;
                    }
                };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<std::string, D>
        {
            static void
            transform(avi::core::ValueOption<std::string, VisualOptionTransformer>& option,
                      godot::Control* root,
                      godot::Node* receiver,
                      std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::LineEdit>(
                    root, "res://scenes/options/string.tscn", option.getLabel(), "String");
                values->connect("text_changed", receiver, "_change");

                values->set_text(option.getValue().c_str());

                std::function<void()> fn = [values, &option]()
                { option.setValue(values->get_text().ascii().get_data()); };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<std::filesystem::path, D>
        {
            static void
            transform(
                avi::core::ValueOption<std::filesystem::path, VisualOptionTransformer>& option,
                godot::Control* root,
                godot::Node* receiver,
                std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::LineEdit>(
                    root, "res://scenes/options/path.tscn", option.getLabel(), "Path");
                values->connect("text_changed", receiver, "_change");

                values->set_text(option.getValue().c_str());

                std::function<void()> fn = [values, &option]()
                { option.setValue(values->get_text().ascii().get_data()); };

                fns.push_back(fn);
            }
        };

    private:
        godot::Control* currentRoot{nullptr};
        godot::Node* currentReceiver{nullptr};
        std::vector<std::function<void()>> returnFunctions;
    };

    template <typename V>
    void
    VisualOptionTransformer::transform(core::ValueOption<V, VisualOptionTransformer>& option)
    {
        Transform<V>::transform(option, currentRoot, currentReceiver, returnFunctions);
    }
} // namespace avi::gui
