#include "InteractionDispatch.h"

#include <cassert>
#include <iostream>

#include <GlobalConstants.hpp>
#include <Input.hpp>
#include <InputEventMouseButton.hpp>

void
avi::interaction::InteractionDispatch::_register_methods()
{
    register_method("_unhandled_input", &InteractionDispatch::_unhandled_input);
    register_method("_process", &InteractionDispatch::_process);
    register_method("_notification", &InteractionDispatch::_notification);
}

void
avi::interaction::InteractionDispatch::_init()
{
    this->manipulator = std::make_unique<Manipulator>(*this);
}

void
avi::interaction::InteractionDispatch::_notification(int what)
{
    if (what == godot::Object::NOTIFICATION_PREDELETE)
    {
        manipulator->preDelete();
    }
}

void
avi::interaction::InteractionDispatch::setup(avi::connection::InteractionSink& interactionSink,
                                             avi::scene::CameraSet& cameraSet,
                                             avi::core::InterfaceUtility& userInterface,
                                             avi::connection::StorageModeSwitch& storageMode)
{
    this->interactionSink = &interactionSink;
    this->cameraSet = &cameraSet;
    this->userInterface = &userInterface;
    this->storageMode = &storageMode;
    avi::interaction::Move* move =
        get_node<avi::interaction::Move>("/root/Root/3D/3D/Visualized Scene/Move");
    avi::interaction::Rotate* rotate =
        get_node<avi::interaction::Rotate>("/root/Root/3D/3D/Visualized Scene/Rotate");
    avi::interaction::Scale* scale =
        get_node<avi::interaction::Scale>("/root/Root/3D/3D/Visualized Scene/Scale");
    avi::interaction::None* none =
        get_node<avi::interaction::None>("/root/Root/3D/3D/Visualized Scene/None");
    this->manipulator->setGizmos(*move, *rotate, *scale, *none);
}

void
avi::interaction::InteractionDispatch::sendInteraction(avi::connection::Interaction& interaction)
{
    interactionSink->addInteraction(interaction);
}

void
avi::interaction::InteractionDispatch::sendManipulation(avi::connection::Interaction& interaction)
{
    interactionSink->addInteraction(interaction);
}

avi::interaction::Manipulator*
avi::interaction::InteractionDispatch::getManipulator()
{
    return this->manipulator.get();
}

void
avi::interaction::InteractionDispatch::_process(float delta)
{
    activeCamera = cameraSet->getActiveCamera();
    godot::Vector2 mousePosition = get_viewport()->get_mouse_position();
    auto target = activeCamera->getSelectedElement(mousePosition);

    bool hasHitSelectable = false;

    if (target.has_value() && target->first)
    {
        avi::core::Element* element = target->first;
        hasHitSelectable = element->isSelectable() || element->hasContextMenu();
    }



    godot::Input* input = godot::Input::get_singleton();

    if (hasHitSelectable)
    {
        input->set_default_cursor_shape(godot::Input::CURSOR_POINTING_HAND);
    }
    else
    {
        input->set_default_cursor_shape(godot::Input::CURSOR_ARROW);
    }
}

void
avi::interaction::InteractionDispatch::_unhandled_input(godot::InputEvent* event)
{
    if (storageMode->isLive())
    {
        auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event);
        if (mouseButtonEvent)
        {
            auto idx = mouseButtonEvent->get_button_index();
            assert(this->cameraSet != nullptr);
            activeCamera = cameraSet->getActiveCamera();
            assert(this->activeCamera != nullptr);

            if (idx == godot::GlobalConstants::BUTTON_RIGHT && event->is_pressed())
            {
                startInteraction();
            }

            if (idx == godot::GlobalConstants::BUTTON_LEFT && event->is_pressed())
            {
                startManipulation();
            }
        }
    }
}

void
avi::interaction::InteractionDispatch::startInteraction()
{
    if (previousElement != nullptr)
    {
        deselect();
    }

    godot::Vector2 mousePosition = get_viewport()->get_mouse_position();
    this->targetElement = activeCamera->getSelectedElement(mousePosition);

    if (!targetElement.has_value())
    {
        return;
    }

    if (targetElement->first == nullptr || targetElement->second == nullptr)
    {
        return;
    }

    if (targetElement->first->hasContextMenu() && targetElement->first->getLayer()->isVisible())
    {
        enableManipulation = false;
        wasManipulation = false;
        std::vector<std::string> interactions = targetElement->first->getInteractions();
        userInterface->openSelectionMenu(mousePosition,
                                         interactions,
                                         [this](int64_t i)
                                         {
                                             this->interaction =
                                                 avi::connection::Interaction::createOption(
                                                     targetElement->first, i);
                                             sendInteraction(interaction);
                                         });
    }
}

void
avi::interaction::InteractionDispatch::startManipulation()
{
    if (!enableManipulation)
    {
        enableManipulation = true;
        return;
    }

    godot::Vector2 mousePosition = get_viewport()->get_mouse_position();
    bool anyHit = false;
    this->targetElement = activeCamera->getSelectedElement(mousePosition, anyHit);

    if (!targetElement.has_value())
    {
        if (!anyHit)
            deselect();
        return;
    }

    if (targetElement->first->isSelectable() && targetElement->first->getLayer()->isVisible())
    {
        deselect();
        this->interaction = avi::connection::Interaction::createSelection(targetElement->first);
        sendInteraction(interaction);
        previousElement = targetElement->first;
        if (targetElement->first->isManipulable())
        {
            this->manipulator->target(
                dynamic_cast<avi::scene::ElementNode&>(*targetElement->first));
            wasManipulation = true;
        }
    }
    else
    {
        deselect();
    }
}

void
avi::interaction::InteractionDispatch::deselect()
{
    if (previousElement)
    {
        avi::connection::Interaction interaction =
            avi::connection::Interaction::createDeselection(previousElement);
        dynamic_cast<avi::scene::ElementNode*>(previousElement)->deselect();

        sendInteraction(interaction);
        previousElement = nullptr;
    }

    if (wasManipulation)
    {
        this->manipulator->untarget();
        wasManipulation = false;
    }
}
