#include "MoveAxisGizmo.h"

#include <sstream>

#include <InputEventMouseButton.hpp>
#include <Viewport.hpp>
#include <World.hpp>
#include <arviz_godot/lib/avi/core/Utilities.h>
#include <arviz_godot/lib/avi/interaction/Move.h>
#include <arviz_godot/lib/avi/interaction/godot/Manipulator.h>

avi::interaction::MoveAxisGizmo::MoveAxisGizmo()
{
}

avi::interaction::MoveAxisGizmo::~MoveAxisGizmo()
{
}

void
avi::interaction::MoveAxisGizmo::_register_methods()
{
    register_method("_input", &MoveAxisGizmo::_input);
    register_method("_process", &MoveAxisGizmo::_process);
    register_method("on_input_event", &MoveAxisGizmo::on_input_event);
}

void
avi::interaction::MoveAxisGizmo::_input(godot::InputEvent* event)
{
    if (auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event))
    {
        auto idx = mouseButtonEvent->get_button_index();
        if (idx == 1 and not(mouseButtonEvent->is_pressed()))
        {
            if (moved)
            {
                move->updateRelease();
            }
            moved = false;
        }
    }
}

void
avi::interaction::MoveAxisGizmo::on_input_event(godot::Object* camera,
                                                godot::InputEvent* event,
                                                godot::Vector3,
                                                godot::Vector3,
                                                int64_t)
{
    this->camera = cast_to<godot::Camera>(camera);

    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event);
    if (mouseButtonEvent and mouseButtonEvent->get_button_index() == 1)
    {
        bool pressed = mouseButtonEvent->is_pressed();

        if (moved and not pressed)
        {
            move->updateRelease();
        }
        else if (not moved and pressed)
        {
            data = godot::Interaction::compute(*mouseButtonEvent,
                                               *this->move,
                                               *this->move->get_proxy(),
                                               *this->camera,
                                               this->axis);
        }
        moved = pressed;
    }
}

void
avi::interaction::MoveAxisGizmo::_process(float)
{
    if (not moved)
    {
        return;
    }

    godot::Vector2 mouse_pos = get_viewport()->get_mouse_position();
    godot::Vector3 mouse_ray_normal_global =
        godot::Interaction::get_ray_normal(*camera, mouse_pos).normalized();
    godot::Vector3 mouse_ray_origin_global = godot::Interaction::get_ray_origin(*camera, mouse_pos);

    auto update = [&](godot::Interaction& _edit)
    {
        using namespace godot;

        auto _get_global_camera_normal = [this]() -> Vector3
        { return -this->camera->get_global_transform().basis.get_axis(2); };

        auto get_global_gizmo_transform = [&]() -> Transform { return _edit.original_global; };

        Vector3 motion_mask;
        Plane plane_global;

        switch (_edit.plane)
        {
            case TRANSFORM_X_AXIS:
                motion_mask = get_global_gizmo_transform().basis.get_axis(0);
                plane_global = Plane(
                    _edit.center_global,
                    motion_mask.cross(motion_mask.cross(_get_global_camera_normal())).normalized());
                break;
            case TRANSFORM_Y_AXIS:
                motion_mask = get_global_gizmo_transform().basis.get_axis(1);
                plane_global = Plane(
                    _edit.center_global,
                    motion_mask.cross(motion_mask.cross(_get_global_camera_normal())).normalized());
                break;
            case TRANSFORM_Z_AXIS:
                motion_mask = get_global_gizmo_transform().basis.get_axis(2);
                plane_global = Plane(
                    _edit.center_global,
                    motion_mask.cross(motion_mask.cross(_get_global_camera_normal())).normalized());
                break;
        }

        Vector3 intersection_global;
        if (not plane_global.intersects_ray(
                mouse_ray_origin_global, mouse_ray_normal_global, &intersection_global))
        {
            return;
        }

        Vector3 click_global;
        if (not plane_global.intersects_ray(
                _edit.click_ray_origin_global, _edit.click_ray_normal_global, &click_global))
        {
            return;
        }

        Vector3 motion_global = intersection_global - click_global;
        motion_global = motion_mask.dot(motion_global) * motion_mask;

        Transform new_global = _edit.proxy_original_global;
        new_global.origin += motion_global;

        move->updateTransform(new_global);
    };

    update(*this->data);
}

void
avi::interaction::MoveAxisGizmo::setup(avi::interaction::Move& move, int axis)
{
    this->move = &move;
    this->axis = axis;

    if (not manipulator)
    {
        manipulator.reset(new godot::Manipulator);
        godot::Manipulator::AxisData axis_data = manipulator->make_axis_data(axis);
        manipulator->init_move_gizmo(axis_data);

        godot::MeshInstance* mesh = godot::MeshInstance::_new();
        mesh->set_mesh(manipulator->move_gizmo[axis]);
        this->add_child(mesh);
    }
}

void
avi::interaction::MoveAxisGizmo::_init()
{
    this->set_capture_input_on_drag(true);
    connect("input_event", this, "on_input_event");
}

void
avi::interaction::MoveAxisGizmo::setActive(bool activeStatus)
{
    if (activeStatus)
    {
        this->show();
    }
    else
    {
        this->hide();
    }
}
