#include "ScaleAxisGizmo.h"

#include <memory>

#include <InputEventMouseButton.hpp>
#include <Math.hpp>
#include <Vector3.hpp>
#include <Viewport.hpp>
#include <arviz_godot/lib/avi/core/Utilities.h>
#include <arviz_godot/lib/avi/interaction/Scale.h>
#include <arviz_godot/lib/avi/interaction/godot/Manipulator.h>

void
avi::interaction::ScaleAxisGizmo::_register_methods()
{
    register_method("_input", &ScaleAxisGizmo::_input);
    register_method("_process", &ScaleAxisGizmo::_process);
    register_method("on_input_event", &ScaleAxisGizmo::on_input_event);
}

avi::interaction::ScaleAxisGizmo::ScaleAxisGizmo()
{
}

avi::interaction::ScaleAxisGizmo::~ScaleAxisGizmo() = default;

void
avi::interaction::ScaleAxisGizmo::_init()
{
    this->set_capture_input_on_drag(true);
    connect("input_event", this, "on_input_event");
}

void
avi::interaction::ScaleAxisGizmo::_input(godot::InputEvent* event)
{
    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event);
    if (mouseButtonEvent)
    {
        auto idx = mouseButtonEvent->get_button_index();
        if (idx == 1 && !(mouseButtonEvent->is_pressed()))
        {
            if (moved)
                scale->updateRelease();
            moved = false;
        }
    }
}

void
avi::interaction::ScaleAxisGizmo::on_input_event(godot::Object* camera,
                                                 godot::InputEvent* event,
                                                 godot::Vector3 position,
                                                 godot::Vector3 normal,
                                                 int64_t shape_idx)
{
    this->camera = cast_to<godot::Camera>(camera);

    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event);
    if (mouseButtonEvent and mouseButtonEvent->get_button_index() == 1)
    {
        bool pressed = mouseButtonEvent->is_pressed();

        if (moved and not pressed)
        {
            scale->updateRelease();
        }
        else if (not moved and pressed)
        {
            data = godot::Interaction::compute(*mouseButtonEvent,
                                               *this->scale,
                                               *this->scale->get_proxy(),
                                               *this->camera,
                                               this->axis);
        }
        moved = pressed;
    }
}

void
avi::interaction::ScaleAxisGizmo::_process(float delta)
{
    if (not moved)
        return;

    godot::Vector2 mouse_pos = get_viewport()->get_mouse_position();
    godot::Vector3 mouse_ray_normal_global =
        godot::Interaction::get_ray_normal(*camera, mouse_pos).normalized();
    godot::Vector3 mouse_ray_origin_global = godot::Interaction::get_ray_origin(*camera, mouse_pos);

    auto update = [&](godot::Interaction& _edit)
    {
        using namespace godot;

        auto _get_global_camera_normal = [this]() -> Vector3
        { return -this->camera->get_global_transform().basis.get_axis(2); };

        auto get_global_gizmo_transform = [&]() -> Transform { return _edit.original_global; };

        Vector3 motion_mask;
        Plane plane;

        switch (_edit.plane)
        {
            case TRANSFORM_X_AXIS:
                motion_mask = get_global_gizmo_transform().basis.get_axis(0).normalized();
                plane = Plane(
                    _edit.center_global,
                    motion_mask.cross(motion_mask.cross(_get_global_camera_normal())).normalized());
                break;
            case TRANSFORM_Y_AXIS:
                motion_mask = get_global_gizmo_transform().basis.get_axis(1).normalized();
                plane = Plane(
                    _edit.center_global,
                    motion_mask.cross(motion_mask.cross(_get_global_camera_normal())).normalized());
                break;
            case TRANSFORM_Z_AXIS:
                motion_mask = get_global_gizmo_transform().basis.get_axis(2).normalized();
                plane = Plane(
                    _edit.center_global,
                    motion_mask.cross(motion_mask.cross(_get_global_camera_normal())).normalized());
                break;
        }

        Vector3 intersection;
        if (!plane.intersects_ray(mouse_ray_origin_global, mouse_ray_normal_global, &intersection))
        {
            return;
        }

        Vector3 click;
        if (!plane.intersects_ray(
                _edit.click_ray_origin_global, _edit.click_ray_normal_global, &click))
        {
            return;
        }

        Vector3 motion = intersection - click;
        motion = motion_mask.dot(motion) * motion_mask;

        motion /= click.distance_to(_edit.center_global);
        motion = _edit.original_global.basis.inverse().xform(motion);

        auto compute_transform = [&](const godot::Transform& original,
                                     const godot::Transform& original_local,
                                     bool is_local) -> Transform
        {
            Transform s;

            if (is_local)
            {
                s.basis = avi::core::utilities::scaled_local(original_local.basis,
                                                             motion + Vector3(1, 1, 1));
                s.origin = original_local.origin;
            }
            else
            {
                s.basis = original_local.basis.scaled(motion + Vector3(1, 1, 1));
                s.origin = original_local.origin;
            }

            return s;
        };

        Transform new_global =
            compute_transform(_edit.proxy_original_global, _edit.proxy_original_local, false);
        Transform new_local =
            compute_transform(_edit.proxy_original_global, _edit.proxy_original_local, true);

        scale->updateTransform(new_global, new_local);
    };

    update(*this->data);
}

void
avi::interaction::ScaleAxisGizmo::setup(avi::interaction::Scale& scale, int axis)
{
    this->scale = &scale;
    this->axis = axis;

    if (not manipulator)
    {
        manipulator = std::make_unique<godot::Manipulator>();
        godot::Manipulator::AxisData axis_data = manipulator->make_axis_data(axis);
        manipulator->init_scale_gizmo(axis_data);

        godot::MeshInstance* mesh = godot::MeshInstance::_new();
        mesh->set_mesh(manipulator->scale_gizmo[axis]);
        this->add_child(mesh);
    }
}

void
avi::interaction::ScaleAxisGizmo::setActive(bool activeStatus)
{
    if (activeStatus)
    {
        this->show();
    }
    else
    {
        this->hide();
    }
}
