#include "ArrowElementNode.h"

#include <CubeMesh.hpp>
#include <CylinderMesh.hpp>
#include <SpatialMaterial.hpp>
#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

avi::scene::ArrowElementNode::ArrowElementNode()
{
}

godot::AABB
avi::scene::ArrowElementNode::getVolume()
{
    godot::AABB correctedAABB = cone->get_transformed_aabb();
    correctedAABB.position = staticBody->get_transform().origin;
    correctedAABB.size = correctedAABB.size * 0.5f;


    return correctedAABB;
}

void
avi::scene::ArrowElementNode::_register_methods()
{
    register_method("_ready", &ArrowElementNode::_ready);
}

void
avi::scene::ArrowElementNode::_ready()
{
}

void
avi::scene::ArrowElementNode::setPosition(godot::Vector3 position)
{
    auto transform = staticBody->get_transform();
    transform.set_origin(position);
    staticBody->set_transform(transform);
}

void
avi::scene::ArrowElementNode::setScale(godot::Vector3 scale)
{
    staticBody->set_scale(scale);
}

void
avi::scene::ArrowElementNode::setRotation(godot::Quat rotation)
{
    godot::Transform quaternionTransform = godot::Transform(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

void
avi::scene::ArrowElementNode::setColor(godot::Color color)
{
    auto* mat = cast_to<godot::SpatialMaterial>(cylinder->get_surface_material(0).ptr());
    mat->set_albedo(color);
    if (color.a < 1)
    {
        mat->set_feature(godot::SpatialMaterial::FEATURE_TRANSPARENT, true);
    }
}

void
avi::scene::ArrowElementNode::setLength(float length)
{
    godot::Vector3 scale = cylinder->get_scale();
    scale.y = length;
    cylinder->set_scale(scale);
    cylinder->set_translation({0, length / 2, 0});
    cone->set_translation({0, length, 0});

    collisionShape->set_translation({0, length / 2, 0});
    collisionShape->set_scale(scale);
}

void
avi::scene::ArrowElementNode::setWidth(float width)
{
    cone->set_scale({width * 2.5f, width * 6.0f, width * 2.5f});
    cylinder->set_scale({width, std::max(0.0f, cylinder->get_scale().y - width * 6.0f), width});
    cylinder->set_translation({0, cylinder->get_scale().y / 2, 0});
    cone->set_translation({0, cylinder->get_scale().y, 0});

    collisionShape->set_translation({0, cylinder->get_scale().y / 2, 0});
    collisionShape->set_scale({width, std::max(0.0f, cylinder->get_scale().y), width});
}

godot::Vector3
avi::scene::ArrowElementNode::getPosition()
{
    return staticBody->get_translation();
}

godot::Vector3
avi::scene::ArrowElementNode::getRotation()
{
    return staticBody->get_rotation();
}

godot::Vector3
avi::scene::ArrowElementNode::getScale()
{
    return staticBody->get_scale() * 1000;
}

avi::scene::ArrowElementNode*
avi::scene::ArrowElementNode::newArrowElementNode(avi::connection::SentElementType type,
                                                  std::string name,
                                                  avi::core::Layer* layer,
                                                  int flag,
                                                  std::vector<std::string> interactions,
                                                  godot::Vector3 position,
                                                  godot::Vector3 scale,
                                                  godot::Quat rotation,
                                                  godot::Color color,
                                                  float length,
                                                  float width)
{
    ArrowElementNode* node = newArrowElementNode();
    node->setAllAttributes(type, name, layer);

    return node;
}

avi::scene::ArrowElementNode*
avi::scene::ArrowElementNode::newArrowElementNode()
{
    ArrowElementNode* node = ArrowElementNode::_new();
    node->staticBody = godot::StaticBody::_new();
    node->collisionShape = godot::CollisionShape::_new();
    node->cylinder = godot::MeshInstance::_new();
    node->cone = godot::MeshInstance::_new();
    node->cylinderMesh = godot::CylinderMesh::_new();
    godot::CylinderMesh* coneMesh = godot::CylinderMesh::_new();
    godot::SpatialMaterial* material = godot::SpatialMaterial::_new();

    node->add_child(node->staticBody);
    node->staticBody->add_child(node->collisionShape);
    node->staticBody->add_child(node->cylinder);
    node->staticBody->add_child(node->cone);

    material->set_flag(godot::SpatialMaterial::FEATURE_TRANSPARENT, true);
    material->set_flag(godot::SpatialMaterial::FLAG_UNSHADED, false);

    node->cylinder->set_mesh(node->cylinderMesh);
    node->cone->set_mesh(coneMesh);
    coneMesh->set_top_radius(0);
    coneMesh->set_height(1);

    node->cylinderMesh->set_height(1);
    node->cylinder->set_scale({0.5, 1, 0.5});
    node->cone->set_scale({1, 1, 1});
    node->cylinder->set_surface_material(0, material);
    node->cone->set_surface_material(0, material);

    return node;
}

void
avi::scene::ArrowElementNode::setAllAttributes(avi::connection::SentElementType type,
                                               std::string name,
                                               avi::core::Layer* layer)
{
    setType(type);
    setName(name);
    setLayer(layer);
}

void
avi::scene::ArrowElementNode::updateAttributes(int flag,
                                               std::vector<std::string> interactions,
                                               godot::Vector3 position,
                                               godot::Vector3 scale,
                                               godot::Quat rotation,
                                               godot::Color color,
                                               float length,
                                               float width)
{
    setInteractionFlag(flag);
    if (!shapeChecked)
    {
        if (ElementNode::isSelectable())
        {
            setShape();
        }
        shapeChecked = true;
    }
    setInteractions(interactions);
    setPosition(position);
    setRotation(rotation);
    setScale(scale);
    setColor(color);
    setLength(length);
    setWidth(width);
}

void
avi::scene::ArrowElementNode::_init()
{
}

void
avi::scene::ArrowElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    auto arrow = IceInternal::Handle<armarx::viz::data::ElementArrow>::dynamicCast(elementInfo);
    updateAttributes(arrow->interaction.enableFlags,
                     arrow->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(arrow->pose),
                     avi::scene::Converter::getScale(arrow->scale),
                     avi::scene::Converter::getRotation(arrow->pose),
                     avi::scene::Converter::getColor(arrow->color),
                     arrow->length,
                     arrow->width);
}

godot::Transform
avi::scene::ArrowElementNode::getTransform()
{
    godot::Transform transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3::ONE * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    return transform;
}

void
avi::scene::ArrowElementNode::setShape()
{
    godot::Ref<godot::Shape> cylinderShape = cylinderMesh->create_trimesh_shape();
    this->collisionShape->set_shape(cylinderShape);
}

std::vector<godot::Spatial*>
avi::scene::ArrowElementNode::getSpatial()
{
    std::vector<godot::Spatial*> vec;
    vec.push_back(cone);
    vec.push_back(cylinder);
    return vec;
}
