#include "LineElementNode.h"

#include <CylinderMesh.hpp>
#include <CylinderShape.hpp>
#include <SphereMesh.hpp>
#include <SphereShape.hpp>
#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/core/Utilities.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

void
avi::scene::LineElementNode::_register_methods()
{
    register_method("_ready", &LineElementNode::_ready);
}

void
avi::scene::LineElementNode::_ready()
{
}

void
avi::scene::LineElementNode::_init()
{
}

void
avi::scene::LineElementNode::setStartAttributes(std::string name, avi::core::Layer* layer)
{
    setType(avi::connection::SentElementType::LINE);
    setName(name);
    setLayer(layer);

    staticBody = godot::StaticBody::_new();
    this->add_child(staticBody);

    start = godot::MeshInstance::_new();
    staticBody->add_child(start);
    end = godot::MeshInstance::_new();
    staticBody->add_child(end);

    godot::Mesh* meshSphere = godot::SphereMesh::_new();
    start->set_mesh(meshSphere);
    end->set_mesh(meshSphere);

    line = godot::MeshInstance::_new();
    staticBody->add_child(line);
    godot::Mesh* meshLine = godot::CylinderMesh::_new();
    line->set_mesh(meshLine);

    material = godot::SpatialMaterial::_new();
    line->set_surface_material(0, material);
    start->set_surface_material(0, material);
    end->set_surface_material(0, material);

    collisionLine = godot::CollisionShape::_new();
    collisionLine->set_shape(godot::CylinderShape::_new());
    staticBody->add_child(collisionLine);
}

void
avi::scene::LineElementNode::updateAttributes(int flag,
                                              std::vector<std::string> interactions,
                                              godot::Vector3 position,
                                              godot::Vector3 scale,
                                              godot::Quat rotation,
                                              godot::Color color)
{
    setInteractionFlag(flag);
    setInteractions(std::move(interactions));

    setPosition(position);
    setRotation(rotation);
    setScale(scale);
    setColor(color);
}

void
avi::scene::LineElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    auto lineHandle = IceInternal::Handle<armarx::viz::data::ElementLine>::dynamicCast(elementInfo);

    if (!godot::Math::is_equal_approx(radius, lineHandle->lineWidth))
    {
        updateRadius(lineHandle->lineWidth);
    }

    godot::Vector3 from = {lineHandle->from.e0, lineHandle->from.e1, lineHandle->from.e2};
    godot::Vector3 to = {lineHandle->to.e0, lineHandle->to.e1, lineHandle->to.e2};

    if (!avi::core::utilities::is_equal_approx(a, from) ||
        !avi::core::utilities::is_equal_approx(b, to))
    {
        updateLine(from, to);
    }

    updateAttributes(lineHandle->interaction.enableFlags,
                     lineHandle->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(lineHandle->pose),
                     avi::scene::Converter::getScale(lineHandle->scale),
                     avi::scene::Converter::getRotation(lineHandle->pose),
                     avi::scene::Converter::getColor(lineHandle->color));
}

avi::scene::LineElementNode*
avi::scene::LineElementNode::newLineElementNode(std::string name, avi::core::Layer* layer)
{
    LineElementNode* node = LineElementNode::_new();
    node->setStartAttributes(name, layer);
    return node;
}

godot::AABB
avi::scene::LineElementNode::getVolume()
{
    godot::AABB aabb(a / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR, godot::Vector3::ZERO);
    aabb.expand_to(b / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    aabb.position += staticBody->get_transform().origin;
    aabb.size *= staticBody->get_scale() * 0.5 * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR;

    return aabb;
}

godot::Transform
avi::scene::LineElementNode::getTransform()
{
    godot::Transform transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3::ONE * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    return transform;
}

std::vector<godot::Spatial*>
avi::scene::LineElementNode::getSpatial()
{
    std::vector<godot::Spatial*> spatial;

    spatial.push_back(start);
    spatial.push_back(end);
    spatial.push_back(line);

    return spatial;
}

void
avi::scene::LineElementNode::updateRadius(float newRadius)
{
    radius = newRadius;

    godot::Vector3 scale = {newRadius, line->get_scale().y, newRadius};
    line->set_scale(scale);
    collisionLine->set_scale(scale);

    float sphereRadius = radius * 1.5f;
    godot::Vector3 sphereScale = {sphereRadius, sphereRadius, sphereRadius};

    start->set_scale(sphereScale);
    end->set_scale(sphereScale);
}

void
avi::scene::LineElementNode::updateLine(godot::Vector3 from, godot::Vector3 to)
{
    a = from;
    b = to;

    start->set_translation(from);
    end->set_translation(to);

    godot::Vector3 direction = to - from;
    godot::Vector3 scale = line->get_scale();
    godot::Vector3 mid = from + direction / 2;

    godot::Quat rotation(godot::Vector3::UP, direction.normalized());
    godot::Transform transform;
    transform.basis *= rotation;
    line->set_transform(transform);

    line->set_translation(mid);
    line->set_scale({scale.x, direction.length() / 2, scale.z});

    collisionLine->set_transform(line->get_transform());
}

void
avi::scene::LineElementNode::setPosition(godot::Vector3 position)
{
    staticBody->set_translation(position);
}

void
avi::scene::LineElementNode::setRotation(godot::Quat rotation)
{
    staticBody->set_rotation(rotation.get_euler());
}

void
avi::scene::LineElementNode::setScale(godot::Vector3 scale)
{
    staticBody->set_scale(scale);
}

void
avi::scene::LineElementNode::setColor(godot::Color color)
{
    material->set_albedo(color);

    if (color.a < 1)
    {
        material->set_feature(godot::SpatialMaterial::FEATURE_TRANSPARENT, true);
    }
}