#include "MeshElementNode.h"

#include <iostream>

#include <RobotAPI/interface/ArViz/Elements.h>

#include <ArrayMesh.hpp>
#include <Shape.hpp>
#include <SpatialMaterial.hpp>
#include <arviz_godot/lib/avi/core/Constants.h>

avi::scene::MeshElementNode::MeshElementNode()
{
}

godot::AABB
avi::scene::MeshElementNode::getVolume()
{
    godot::AABB correctedAABB = meshInstance->get_aabb();
    correctedAABB.position = staticBody->get_global_transform().origin;
    correctedAABB.size *= meshInstance->get_scale() * staticBody->get_scale() * 0.5;

    return correctedAABB;
}

void
avi::scene::MeshElementNode::setPosition(godot::Vector3 position)
{
    auto transform = staticBody->get_transform();
    transform.set_origin(position);

    staticBody->set_transform(transform);
}

void
avi::scene::MeshElementNode::setScale(godot::Vector3 scale)
{
    staticBody->set_scale(scale);
}

void
avi::scene::MeshElementNode::setRotation(godot::Quat rotation)
{
    godot::Transform quaternionTransform = godot::Transform(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

void
avi::scene::MeshElementNode::setColor(godot::Color color)
{
    auto* mat = cast_to<godot::SpatialMaterial>(meshInstance->get_surface_material(0).ptr());
    mat->set_albedo(color);
    if (color.a < 1)
    {
        mat->set_feature(godot::SpatialMaterial::FEATURE_TRANSPARENT, true);
    }
}

godot::Vector3
avi::scene::MeshElementNode::getPosition()
{
    return staticBody->get_translation();
}

godot::Vector3
avi::scene::MeshElementNode::getRotation()
{
    return staticBody->get_rotation();
}

godot::Vector3
avi::scene::MeshElementNode::getScale()
{
    return staticBody->get_scale();
}

void
avi::scene::MeshElementNode::setStartAttributes(godot::Mesh* mesh,
                                                avi::connection::SentElementType type,
                                                std::string name,
                                                avi::core::Layer* layer)
{
    staticBody = godot::StaticBody::_new();
    collisionShape = godot::CollisionShape::_new();
    meshInstance = godot::MeshInstance::_new();

    add_child(staticBody);
    staticBody->add_child(collisionShape);

    staticBody->add_child(meshInstance);

    godot::SpatialMaterial* material = godot::SpatialMaterial::_new();
    material->set_flag(godot::SpatialMaterial::FLAG_UNSHADED, false);

    meshInstance->set_mesh(godot::Ref<godot::Mesh>{mesh});
    meshInstance->set_surface_material(0, material);

    setMesh(mesh);
    setType(type);
    setName(name);
    setLayer(layer);
}

void
avi::scene::MeshElementNode::updateAttributes(int flag,
                                              std::vector<std::string> interactions,
                                              godot::Vector3 position,
                                              godot::Vector3 scale,
                                              godot::Quat rotation,
                                              godot::Color color)
{
    setInteractionFlag(flag);
    setInteractions(interactions);
    setPosition(position);
    setRotation(rotation);
    setScale(scale);
    setColor(color);
}

godot::Mesh*
avi::scene::MeshElementNode::getMesh()
{
    return meshInstance->get_mesh().ptr();
}

void
avi::scene::MeshElementNode::setMesh(godot::Mesh* mesh)
{
    meshInstance->set_mesh(godot::Ref<godot::Mesh>{mesh});
}

void
avi::scene::MeshElementNode::useVertexColor(bool isUsed)
{
    auto* mat = cast_to<godot::SpatialMaterial>(meshInstance->get_surface_material(0).ptr());
    mat->set_flag(godot::SpatialMaterial::FLAG_ALBEDO_FROM_VERTEX_COLOR, isUsed);
}

void
avi::scene::MeshElementNode::setTwoSided()
{
    auto* mat = cast_to<godot::SpatialMaterial>(meshInstance->get_surface_material(0).ptr());
    mat->set_cull_mode(godot::SpatialMaterial::CULL_DISABLED);
}

void
avi::scene::MeshElementNode::setShape(godot::Mesh* mesh)
{
    godot::Ref<godot::Shape> shape = mesh->create_trimesh_shape();
    this->collisionShape->set_shape(shape);
    collisionShape->set_scale(meshInstance->get_scale());
}

void
avi::scene::MeshElementNode::setShape(godot::Shape* shape)
{
    this->collisionShape->set_shape(shape);
}

godot::Transform
avi::scene::MeshElementNode::getTransform()
{
    godot::Transform transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3::ONE * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    return transform;
}

std::vector<godot::Spatial*>
avi::scene::MeshElementNode::getSpatial()
{
    std::vector<godot::Spatial*> vec;
    vec.push_back(meshInstance);
    return vec;
}

godot::MeshInstance*
avi::scene::MeshElementNode::getMeshInstance()
{
    return meshInstance;
}

godot::CollisionShape*
avi::scene::MeshElementNode::getCollisionShape()
{
    return collisionShape;
}
