#include "ObjectElementNode.h"

#include <iostream>

#include <VirtualRobot/ManipulationObject.h>

#include <CubeMesh.hpp>
#include <SpatialMaterial.hpp>
#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

avi::scene::ObjectElementNode::ObjectElementNode()
{
}

godot::AABB
avi::scene::ObjectElementNode::getVolume()
{
    if (meshes.empty())
    {
        // Object files can be empty.

        godot::AABB emptyAABB;
        emptyAABB.position = staticBody->get_global_transform().origin;
        emptyAABB.size = godot::Vector3::ZERO;

        return emptyAABB;
    }

    godot::AABB AABB = meshes.front().meshInstance->get_aabb();
    for (size_t i = 1; i < meshes.size(); ++i)
    {
        AABB.merge_with(meshes[i].meshInstance->get_aabb());
    }
    AABB.position = staticBody->get_global_transform().origin;
    AABB.size = AABB.size * staticBody->get_scale() * 0.5;

    return AABB;
}

void
avi::scene::ObjectElementNode::_register_methods()
{
    register_method("_ready", &ObjectElementNode::_ready);
}

void
avi::scene::ObjectElementNode::_ready()
{
}

void
avi::scene::ObjectElementNode::setPosition(godot::Vector3 position)
{
    auto transform = staticBody->get_transform();
    transform.set_origin(position);
    staticBody->set_transform(transform);
}

void
avi::scene::ObjectElementNode::setScale(godot::Vector3 scale)
{
    staticBody->set_scale(scale);
}

void
avi::scene::ObjectElementNode::setRotation(godot::Quat rotation)
{
    godot::Transform quaternionTransform = godot::Transform(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

godot::Vector3
avi::scene::ObjectElementNode::getPosition()
{
    return staticBody->get_translation();
}

godot::Vector3
avi::scene::ObjectElementNode::getRotation()
{
    return staticBody->get_rotation();
}

godot::Vector3
avi::scene::ObjectElementNode::getScale()
{
    return staticBody->get_scale();
}

avi::scene::ObjectElementNode*
avi::scene::ObjectElementNode::newObjectElementNode(avi::core::ModelLoader* loader,
                                                    avi::connection::SentElementType type,
                                                    std::string name,
                                                    avi::core::Layer* layer)
{

    ObjectElementNode* node = ObjectElementNode::_new();
    node->staticBody = godot::StaticBody::_new();
    node->add_child(node->staticBody);
    node->loader = loader;

    node->setStartAttributes(type, name, layer);

    return node;
}

void
avi::scene::ObjectElementNode::setStartAttributes(avi::connection::SentElementType type,
                                                  std::string name,
                                                  avi::core::Layer* layer)
{
    setType(type);
    setName(name);
    setLayer(layer);
}

void
avi::scene::ObjectElementNode::updateAttributes(std::string fileName,
                                                int flag,
                                                std::vector<std::string> interactions,
                                                godot::Vector3 position,
                                                godot::Vector3 scale,
                                                godot::Quat rotation,
                                                godot::Color color,
                                                int drawStyle)
{
    setInteractionFlag(flag);
    if (this->fileName != fileName)
    {
        // Remove old mesh instances and collision shapes.
        {
            for (Mesh& mesh : meshes)
            {
                staticBody->remove_child(mesh.meshInstance);

                staticBody->remove_child(mesh.collisionShape);
            }

            meshes.clear();
        }

        this->fileName = fileName;
        std::string fullName;
        avi::core::Object object = loader->loadModel(fileName, fullName);

        for (core::Mesh& m : object.meshes)
        {
            if (m.material->get_albedo().a != 1)
            {
                m.material->set_feature(godot::SpatialMaterial::FEATURE_TRANSPARENT, true);
            }

            Mesh& mesh = meshes.emplace_back();
            {
                mesh.meshInstance = godot::MeshInstance::_new();
                mesh.meshInstance->set_mesh(godot::Ref<godot::Mesh>{m.mesh});
                staticBody->add_child(mesh.meshInstance);
            }
            {
                mesh.collisionShape = godot::CollisionShape::_new();
                staticBody->add_child(mesh.collisionShape);
            }
        }

        if (!shapeChecked)
        {
            setShape(loader->getObject(fullName));
            shapeChecked = true;
        }
    }

    if (currentDrawStyle != drawStyle || currentColor != color)
    {
        currentDrawStyle = drawStyle;
        currentColor = color;

        setOverrideColor(color, drawStyle);
    }

    setInteractions(interactions);
    setPosition(position);
    setRotation(rotation);
    setScale(scale);
}

void
avi::scene::ObjectElementNode::_init()
{
}

void
avi::scene::ObjectElementNode::useVertexColor(bool isUsed)
{
    for (Mesh& mesh : meshes)
    {
        auto* mat = cast_to<godot::SpatialMaterial>(mesh.meshInstance->get_surface_material(0).ptr());
        mat->set_flag(godot::SpatialMaterial::FLAG_ALBEDO_FROM_VERTEX_COLOR, isUsed);
    }
}

void
avi::scene::ObjectElementNode::setTwoSided()
{
    for (Mesh& mesh : meshes)
    {
        auto* mat = cast_to<godot::SpatialMaterial>(mesh.meshInstance->get_surface_material(0).ptr());
        mat->set_cull_mode(godot::SpatialMaterial::CULL_DISABLED);
    }
}

void
avi::scene::ObjectElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    auto object = IceInternal::Handle<armarx::viz::data::ElementObject>::dynamicCast(elementInfo);
    updateAttributes(object->filename,
                     object->interaction.enableFlags,
                     object->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(object->pose),
                     godot::Vector3(object->scale.e0, object->scale.e1, object->scale.e2),
                     avi::scene::Converter::getRotation(object->pose),
                     avi::scene::Converter::getColor(object->color),
                     object->drawStyle);
}

godot::Transform
avi::scene::ObjectElementNode::getTransform()
{
    godot::Transform transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3::ONE * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);
    return transform;
}

void
avi::scene::ObjectElementNode::setShape(VirtualRobot::ManipulationObjectPtr visualization)
{
    if (visualization && visualization->getCollisionModel())
    {
        core::Object obj = avi::core::ModelLoader::createObjectFromTriMesh(visualization->getCollisionModel()->getTriMeshModel());
        meshes[0].collisionShape->set_shape(obj.meshes.front().mesh->create_trimesh_shape());
    }
    else
    {
        for (Mesh& mesh : meshes)
        {
            std::optional<avi::core::Shape> shape =
                loader->getShapeCache().get(mesh.meshInstance->get_mesh().ptr());

            if (shape)
            {
                mesh.collisionShape->set_shape(shape->shape);
            }
            else
            {
                godot::Ref<godot::Shape> newShape =
                    mesh.meshInstance->get_mesh()->create_trimesh_shape();
                loader->getShapeCache().emplace(mesh.meshInstance->get_mesh().ptr(),
                                                {newShape.ptr()});

                mesh.collisionShape->set_shape(newShape);
            }
        }
    }
}

std::vector<godot::Spatial*>
avi::scene::ObjectElementNode::getSpatial()
{
    std::vector<godot::Spatial*> vec;

    for (Mesh& mesh : meshes)
    {
        vec.push_back(mesh.meshInstance);
    }

    return vec;
}


void
avi::scene::ObjectElementNode::setOverrideColor(godot::Color color, int drawStyle)
{
    std::string fullName;
    avi::core::Object object = loader->loadModel(fileName, fullName);

    if (drawStyle == armarx::viz::data::ModelDrawStyle::OVERRIDE_COLOR)
    {
        godot::Ref<godot::SpatialMaterial> material =
            meshes.empty() || meshes.front().meshInstance->get_material_override().is_null()
                ? godot::SpatialMaterial::_new()
                : meshes.front().meshInstance->get_material_override();

        material->set_albedo(color);
        material->set_flag(godot::SpatialMaterial::FLAG_ALBEDO_FROM_VERTEX_COLOR, true);

        material->set_blend_mode(godot::SpatialMaterial::BLEND_MODE_MIX);
        material->set_feature(godot::SpatialMaterial::FEATURE_TRANSPARENT, color.a != 1);

        int i = 0;
        for (auto& mesh : meshes)
        {
            material->set_texture(
                godot::SpatialMaterial::TEXTURE_ALBEDO,
                object.meshes[i++].material->get_texture(godot::SpatialMaterial::TEXTURE_ALBEDO));
            mesh.meshInstance->set_material_override(material);
        }
    }
    else if (drawStyle == armarx::viz::data::ModelDrawStyle::ORIGINAL)
    {
        for (auto& mesh : meshes)
        {
            mesh.meshInstance->set_material_override(nullptr);
        }
    }
}