#include "PlayfulControlScheme.h"

#include <utility>

#include <Godot.hpp>
#include <Input.hpp>
#include <InputEvent.hpp>
#include <GlobalConstants.hpp>
#include <InputEventMouseMotion.hpp>
#include <InputEventMouseButton.hpp>
#include <Control.hpp>
#include <SceneTree.hpp>

void
avi::scene::PlayfulControlScheme::_register_methods()
{
    register_method("_ready", &PlayfulControlScheme::_ready);
    register_method("_unhandled_input", &PlayfulControlScheme::_unhandled_input);
    register_method("_process", &PlayfulControlScheme::_process);
}

void
avi::scene::PlayfulControlScheme::_init()
{
    viewCamera = nullptr;
    scrollValue = 0.0f;

    inputState = std::vector<bool>(MAX_INPUT_INDEX, false);

    inputKeys.push_back("avi_camera_forward");
    inputKeys.push_back("avi_camera_backward");
    inputKeys.push_back("avi_camera_left");
    inputKeys.push_back("avi_camera_right");
    inputKeys.push_back("avi_camera_up");
    inputKeys.push_back("avi_camera_down");
    inputKeys.push_back("avi_camera_roll_left");
    inputKeys.push_back("avi_camera_roll_right");
}

void
avi::scene::PlayfulControlScheme::setup(avi::core::SettingsProvider& settings)
{
    this->settings = &settings;
}

void
avi::scene::PlayfulControlScheme::_ready()
{
    attachment = get_node<godot::Spatial>("Attachment");
}

void
avi::scene::PlayfulControlScheme::_unhandled_input(godot::InputEvent* event)
{
    if (!isActive()) return;

    auto* mouseMotionEvent = cast_to<godot::InputEventMouseMotion>(event);
    if (mouseMotionEvent)
    {
        mouseDelta = mouseMotionEvent->get_relative();
    }

    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event);
    if (mouseButtonEvent)
    {
        float change = 0.0f;
        auto idx = mouseButtonEvent->get_button_index();

        if (idx == godot::GlobalConstants::BUTTON_WHEEL_UP)
        {
            change += 1.0f;
        }

        if (idx == godot::GlobalConstants::BUTTON_WHEEL_DOWN)
        {
            change -= 1.0f;
        }

        scrollValue += change;
        scrollValue = std::clamp(scrollValue, -100.0f, 200.0f);
    }

    for (auto i = 0; i < MAX_INPUT_INDEX; i++)
    {
        if (event->is_action_pressed(inputKeys[i]))
            inputState[i] = true;
    }
}

void
avi::scene::PlayfulControlScheme::_process(float delta)
{
    if (!isActive()) return;

    godot::Vector2 oldMouseDelta = mouseDelta;
    mouseDelta = {0, 0};

    if (isTraveling)
    {
        doTravel(delta);
        return;
    }

    godot::Input* input = godot::Input::get_singleton();

    for (auto i = 0; i < MAX_INPUT_INDEX; i++)
    {
        if (!input->is_action_pressed(inputKeys[i]))
            inputState[i] = false;
    }

    bool rotate = input->is_action_pressed("avi_camera_mouse") || viewCamera->isMouseLocked();

    float mouseSensitivity = static_cast<float>(settings->getMouseSensitivity()) * 0.001;

    if (rotate)
    {
        attachment->rotate_object_local({0, 1, 0},
                                        mouseSensitivity * oldMouseDelta.x * delta * -1.0f);
        attachment->rotate_object_local({1, 0, 0},
                                        mouseSensitivity * oldMouseDelta.y * delta * -1.0f);
    }

    godot::Vector3 movement;
    if (inputState[INPUT_FORWARD])
        movement += {0, 0, -1};
    if (inputState[INPUT_BACKWARD])
        movement += {0, 0, 1};
    if (inputState[INPUT_LEFT])
        movement += {-1, 0, 0};
    if (inputState[INPUT_RIGHT])
        movement += {1, 0, 0};
    if (inputState[INPUT_UP])
        movement += {0, 1, 0};
    if (inputState[INPUT_DOWN])
        movement += {0, -1, 0};

    if (movement.length() > 0)
    {
        movement.normalize();
        movement *= delta * static_cast<float>(settings->getMoveSpeed() * 0.01);
        movement *= std::pow(1.05f, scrollValue);

        attachment->translate(movement);
    }

    float roll = 0;
    if (inputState[INPUT_ROLL_LEFT])
        roll += 1;
    if (inputState[INPUT_ROLL_RIGHT])
        roll -= 1;

    if (roll != 0)
    {
        roll *= delta * static_cast<float>(settings->getRollSensitivityPlayful() * 0.02);
        attachment->rotate_object_local({0, 0, 1}, roll);
    }
}

void
avi::scene::PlayfulControlScheme::takeOver(avi::scene::ViewCamera* viewCamera)
{
    this->viewCamera = viewCamera;

    attachment->set_transform(viewCamera->get_transform());
    viewCamera->set_transform(godot::Transform());

    attachment->add_child(viewCamera);
}

void
avi::scene::PlayfulControlScheme::release()
{
    attachment->remove_child(viewCamera);
    viewCamera->set_transform(attachment->get_transform());
    this->viewCamera = nullptr;
}

void
avi::scene::PlayfulControlScheme::reset()
{
    startTravel(godot::Transform(), 0.5f);
}

void
avi::scene::PlayfulControlScheme::focus(godot::AABB box)
{
    godot::Transform transform;
    godot::Vector3 aabbCenter = box.position + box.size * 0.5f;

    float zoomValue = box.size.length() * 3;
    godot::Vector3 zoomVector =
        (aabbCenter - attachment->get_global_transform().origin).normalized();

    transform.origin = aabbCenter - zoomVector * zoomValue;
    transform = transform.looking_at(box.position + box.size * 0.5f, {0, 1, 0});

    startTravel(transform, 0.5f);

    float clipPlaneValue = zoomValue * 0.01f;
    this->viewCamera->setClipPlanes(clipPlaneValue);
}

void
avi::scene::PlayfulControlScheme::focus(godot::Vector3 point, avi::core::Element* target)
{
    godot::AABB box = target->getVolume();
    godot::Transform transform;

    float zoomValue = box.size.length() * 3;
    godot::Vector3 zoomVector = (point - attachment->get_global_transform().origin).normalized();

    transform.origin = point - zoomVector * zoomValue;
    transform = transform.looking_at(box.position + box.size * 0.5f, {0, 1, 0});

    startTravel(transform, 0.5f);

    float clipPlaneValue = zoomValue * 0.01f;
    this->viewCamera->setClipPlanes(clipPlaneValue);
}

godot::Basis
avi::scene::PlayfulControlScheme::getRotationBasis()
{
    return this->attachment->get_transform().basis;
}

bool
avi::scene::PlayfulControlScheme::isActive()
{
    return viewCamera != nullptr;
}

void
avi::scene::PlayfulControlScheme::startTravel(godot::Transform target, float duration)
{
    travel = {attachment->get_transform(), std::move(target), duration, duration};
    isTraveling = true;
}

void
avi::scene::PlayfulControlScheme::doTravel(float delta)
{
    travel.remainingTime -= delta;
    float progress = 1.0f - (travel.remainingTime / travel.duration);

    if (travel.remainingTime < 0)
    {
        isTraveling = false;
        progress = 1.0f;
    }

    godot::Transform transform =
        travel.startTransform.interpolate_with(travel.endTransform, progress);
    attachment->set_transform(transform);
}
