#include "RobotElementNode.h"

#include <VirtualRobot/CollisionDetection/CollisionModel.h>
#include <VirtualRobot/VirtualRobot.h>
#include <VirtualRobot/Visualization/VisualizationNode.h>

#include <CubeMesh.hpp>
#include <MainLoop.hpp>
#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/core/Utilities.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

avi::scene::RobotElementNode::RobotElementNode()
{
}
void
avi::scene::RobotElementNode::_register_methods()
{

    register_method("_notification", &RobotElementNode::_notification);
}

godot::AABB
avi::scene::RobotElementNode::getVolume()
{
    bool noBaseAABB = true;
    godot::AABB AABB;
    godot::Vector3 robotScale = staticBody->get_scale() * 0.75;

    for (auto pair : nodes)
    {
        if (noBaseAABB)
        {
            AABB = pair.second->get_aabb();
            AABB.position = pair.second->get_global_transform().origin;
            AABB.size *= robotScale;
            noBaseAABB = false;
        }
        else
        {
            godot::AABB t = pair.second->get_aabb();
            t.position = pair.second->get_global_transform().origin;
            t.size *= robotScale;
            AABB = AABB.merge(t);
        }
    }
    AABB.size *= 0.5;
    AABB.position += AABB.size;

    return AABB;
}

void
avi::scene::RobotElementNode::setPosition(godot::Vector3 position)
{
    auto transform = staticBody->get_transform();
    transform.set_origin(position);
    staticBody->set_transform(transform);
}

void
avi::scene::RobotElementNode::setScale(godot::Vector3 scale)
{
    staticBody->set_scale(scale);
}

void
avi::scene::RobotElementNode::setRotation(godot::Quat rotation)
{
    godot::Transform quaternionTransform = godot::Transform(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

void
avi::scene::RobotElementNode::setStartAttributes(avi::scene::GraphicsController* controller,
                                                 avi::core::ModelLoader* modelLoader,
                                                 avi::connection::SentElementType type,
                                                 std::string name,
                                                 avi::core::Layer* layer,
                                                 int flag,
                                                 std::vector<std::string> interactions,
                                                 godot::Vector3 position,
                                                 godot::Vector3 scale,
                                                 godot::Quat rotation,
                                                 godot::Color color,
                                                 std::string project,
                                                 std::string filename,
                                                 int drawStyle,
                                                 std::map<std::string, float> jointValues)
{
    this->graphicsController = controller;
    this->modelLoader = modelLoader;

    setName(name);
    setLayer(layer);
    setType(type);
    updateAttributes(flag,
                     interactions,
                     position,
                     scale,
                     rotation,
                     color,
                     project,
                     filename,
                     drawStyle,
                     jointValues);

    controller->registerChangeListener(this, [this]() { this->visualizationChanged(); });
}

// Loading and updating robot from RobotAPI/components/ArViz/Coin/VisualizationRobot.cpp
void
avi::scene::RobotElementNode::updateAttributes(int flag,
                                               std::vector<std::string> interactions,
                                               godot::Vector3 position,
                                               godot::Vector3 scale,
                                               godot::Quat rotation,
                                               godot::Color color,
                                               std::string project,
                                               std::string filename,
                                               int drawStyle,
                                               const std::map<std::string, float>& jointValues)
{
    setInteractionFlag(flag);
    setInteractions(interactions);
    setPosition(position);
    setRotation(rotation);
    setScale(scale * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    bool robotChanged = this->project != project || this->filename != filename;
    if (robotChanged || !robot)
    {
        robot = this->loadRobot(project, filename);
        visualizeRobot();
    }
    if (!robot)
    {
        ARMARX_ERROR << "Error loading robot from " << filename;
        return;
    }

    VirtualRobot::Robot& ro = *robot;
    for (auto& [nodeName, newJointValue] : jointValues)
    {
        VirtualRobot::RobotNodePtr robotNode = ro.getRobotNode(nodeName);
        float oldJointValue = robotNode->getJointValue();
        float diff = std::abs(newJointValue - oldJointValue);
        bool jointValuesChanged = diff > 0.001f;
        if (jointValuesChanged)
        {
            // Only set the joint values if they changed.
            // This call causes internal updates to the visualization even if joint angles are the same!
            ro.setJointValues(jointValues);
            break;
        }
    }

    if (this->armarXDrawStyle != drawStyle || this->color != color)
    {
        this->armarXDrawStyle = drawStyle;
        this->color = color;

        isOverrideDirty = true;
    }

    visualizationChanged();
}

void
avi::scene::RobotElementNode::updateVisualization()
{
    VirtualRobot::Robot& ro = *robot;

    for (const auto& nodeName : ro.getRobotNodeNames())
    {
        if (nodes.find(nodeName) != nodes.end())
        {
            auto node = ro.getRobotNode(nodeName);
            auto nodePosition = node->getGlobalPose();

            godot::Transform transform = godot::Transform(
                {{nodePosition(0, 0), nodePosition(0, 1), nodePosition(0, 2)},
                 {nodePosition(1, 0), nodePosition(1, 1), nodePosition(1, 2)},
                 {nodePosition(2, 0), nodePosition(2, 1), nodePosition(2, 2)}},
                {nodePosition(0, 3) / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR,
                 nodePosition(1, 3) / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR,
                 nodePosition(2, 3) / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR});

            nodes.at(nodeName)->set_transform(transform);

            if (collisionShapes.find(nodeName) != collisionShapes.end())
            {
                collisionShapes.at(nodeName)->set_transform(transform);
            }
        }
    }
}

avi::scene::RobotElementNode*
avi::scene::RobotElementNode::newRobotElementNode()
{
    RobotElementNode* node = avi::scene::RobotElementNode::_new();
    node->staticBody = godot::StaticBody::_new();
    node->add_child(node->staticBody);
    return node;
}

VirtualRobot::RobotPtr
avi::scene::RobotElementNode::loadRobot(std::string project, std::string filename)
{
    VirtualRobot::RobotPtr result;

    auto cacheEntry = getCache().get(project, filename);
    if (cacheEntry.has_value())
    {
        return cacheEntry.value();
    }

    if (filename.empty())
    {
        ARMARX_WARNING << "Error loading robot: filename is emtpy";
        return result;
    }

    if (project.empty() && std::filesystem::path(filename).is_absolute())
    {
        ARMARX_INFO << deactivateSpam() << "You specified the absolute path to the robot file:"
                    << "\n\t'" << filename << "'"
                    << "\nConsider specifying the containing ArmarX package and relative data path "
                       "instead to "
                    << "improve portability to other systems.";
    }
    std::string fullFilename;
    armarx::ArmarXDataPath::FindPackageAndAddDataPath(project);

    if (!armarx::ArmarXDataPath::SearchReadableFile(filename, fullFilename))
    {
        ARMARX_WARNING << "Robot could not be loaded";
        return result;
    }

    VirtualRobot::RobotIO::RobotDescription loadMode = VirtualRobot::RobotIO::eFull;

    try
    {
        ARMARX_INFO << "Loading robot from " << fullFilename;
        result = VirtualRobot::RobotIO::loadRobot(fullFilename, loadMode);
        if (result)
        {
            result->setThreadsafe(false);
            result->setPropagatingJointValuesEnabled(false);
        }
        else
        {
            ARMARX_WARNING << "Could not load robot from file: " << fullFilename;
        }
    }
    catch (VirtualRobot::VirtualRobotException& e)
    {
        ARMARX_WARNING << "Could not load robot from file: " << fullFilename
                       << "\nReason: " << e.what();
    }

    this->project = project;
    this->filename = filename;

    getCache().set(project, filename, result);

    return result;
}

void
avi::scene::RobotElementNode::_init()
{
}

void
avi::scene::RobotElementNode::visualizeRobot()
{
    isOverrideDirty = true;

    for (auto node : nodes)
    {
        node.second->queue_free();
    }

    nodes.clear();

    for (const auto& collisionShape : collisionShapes)
    {
        collisionShape.second->queue_free();
    }

    collisionShapes.clear();

    if (!robot)
    {
        return;
    }

    for (const std::string& nodeName : robot->getRobotNodeNames())
    {
        auto node = robot->getRobotNode(nodeName);

        auto loadObject = [&](VirtualRobot::VisualizationNodePtr visualization)
            -> std::optional<avi::core::Object>
        {
            std::string nodeFilename = visualization->getFilename();

            if (not nodeFilename.empty())
            {
                avi::core::Object object = modelLoader->loadModelFromFile(nodeFilename);

                ARMARX_CHECK_POSITIVE(object.meshes.size());
                return object;
            }

            if (visualization->getTriMeshModel())
            {
                avi::core::Object object = avi::core::ModelLoader::createObjectFromTriMesh(
                    visualization->getTriMeshModel());

                ARMARX_CHECK_POSITIVE(object.meshes.size());
                return object;
            }
            return {};
        };

        auto addMeshFromObject = [&](std::optional<avi::core::Object>& object)
        {
            if (not object.has_value())
                return;

            auto actualObject = object.value();
            const avi::core::Mesh& mesh = actualObject.meshes.front();
            godot::MeshInstance* m = godot::MeshInstance::_new();
            m->set_name(nodeName.c_str());
            m->set_mesh(mesh.mesh);

            staticBody->add_child(m);
            nodes.emplace(nodeName, m);
        };

        auto addShapeFromObject = [&](std::optional<avi::core::Object>& object)
        {
            if (not object.has_value())
            {
                return;
            }

            auto actualObject = object.value();

            avi::core::Mesh& mesh = actualObject.meshes.front();
            addShape(mesh.mesh.ptr(), nodeName);
        };

        VirtualRobot::VisualizationNodePtr visualization = node->getVisualization();
        VirtualRobot::VisualizationNodePtr collision =
            node->getCollisionModel() ? node->getCollisionModel()->getVisualization() : nullptr;

        bool useCollisionAsVisualization = drawStyle != avi::scene::DrawStyle::VISUALIZATION;

        if (useCollisionAsVisualization)
        {
            if (collision)
            {
                auto object = loadObject(collision);
                addMeshFromObject(object);
                addShapeFromObject(object);
            }
        }
        else
        {
            if (visualization)
            {
                auto object = loadObject(visualization);
                addMeshFromObject(object);
            }

            if (collision)
            {
                auto object = loadObject(collision);
                addShapeFromObject(object);
            }
        }
    }
}

void
avi::scene::RobotElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    auto robot = IceInternal::Handle<armarx::viz::data::ElementRobot>::dynamicCast(elementInfo);

    updateAttributes(robot->interaction.enableFlags,
                     robot->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(robot->pose),
                     avi::scene::Converter::getScale(robot->scale),
                     avi::scene::Converter::getRotation(robot->pose),
                     avi::scene::Converter::getColor(robot->color),
                     robot->project,
                     robot->filename,
                     robot->drawStyle,
                     robot->jointValues);
}

godot::Transform
avi::scene::RobotElementNode::getTransform()
{
    return staticBody->get_transform();
}

void
avi::scene::RobotElementNode::addShape(godot::Mesh* mesh, const std::string& nodeName)
{
    godot::CollisionShape* collisionShape = godot::CollisionShape::_new();

    std::optional<avi::core::Shape> shape = modelLoader->getShapeCache().get(mesh);

    if (shape)
    {
        collisionShape->set_shape(shape->shape);
    }
    else
    {
        godot::Ref<godot::Shape> newShape = mesh->create_trimesh_shape();
        modelLoader->getShapeCache().emplace(mesh, {newShape.ptr()});

        collisionShape->set_shape(newShape);
    }

    staticBody->add_child(collisionShape);
    collisionShapes.emplace(nodeName, collisionShape);
}

void
avi::scene::RobotElementNode::visualizationChanged()
{
    if (!robot)
    {
        robot = this->loadRobot(project, filename);
    }

    if (!robot)
    {
        ARMARX_ERROR << "Error loading robot from '" << filename << "'";
        return;
    }

    bool drawStyleChanged = false;
    avi::scene::DrawStyle newDrawStyle = graphicsController->getCurrentDrawStyle();
    if (newDrawStyle == avi::scene::DrawStyle::ORIGINAL)
    {
        if (drawStyle == avi::scene::DrawStyle::VISUALIZATION &&
            armarXDrawStyle == armarx::viz::data::ModelDrawStyle::COLLISION)
        {
            drawStyleChanged = true;
            drawStyle = avi::scene::DrawStyle::COLLISION;
        }
        else if (drawStyle == avi::scene::DrawStyle::COLLISION &&
                 (armarXDrawStyle == armarx::viz::data::ModelDrawStyle::ORIGINAL ||
                  armarXDrawStyle == armarx::viz::data::ModelDrawStyle::OVERRIDE_COLOR))
        {
            drawStyleChanged = true;
            drawStyle = avi::scene::DrawStyle::VISUALIZATION;
        }
    }
    else if (newDrawStyle != drawStyle)
    {
        drawStyleChanged = true;
        drawStyle = newDrawStyle;
    }

    if (drawStyleChanged)
    {
        visualizeRobot();
    }

    updateVisualization();
    setOverrideColor(this->color, this->armarXDrawStyle);
}

void
avi::scene::RobotElementNode::_notification(int what)
{
    if (what == godot::MainLoop::NOTIFICATION_PREDELETE)
    {
        graphicsController->unregisterChangeListener(this);
    }
}

std::vector<godot::Spatial*>
avi::scene::RobotElementNode::getSpatial()
{
    std::vector<godot::Spatial*> vec;
    for (auto node : nodes)
    {
        vec.push_back(node.second);
    }
    return vec;
}

void
avi::scene::RobotElementNode::setOverrideColor(godot::Color color, int drawStyle)
{
    if (not isOverrideDirty)
    {
        return;
    }

    if (drawStyle == armarx::viz::data::ModelDrawStyle::OVERRIDE_COLOR)
    {
        if (not nodes.empty())
        {
            auto& [key, value] = *nodes.begin();
            godot::Ref<godot::SpatialMaterial> material;

            if (value->get_material_override().is_null())
            {
                material = godot::Ref(godot::SpatialMaterial::_new());
            }
            else
            {
                material = value->get_material_override();
            }

            material->set_albedo(color);
            material->set_flag(godot::SpatialMaterial::FLAG_ALBEDO_FROM_VERTEX_COLOR, true);

            if (color.a != 1)
            {
                material->set_blend_mode(godot::SpatialMaterial::BLEND_MODE_MIX);
                material->set_feature(godot::SpatialMaterial::FEATURE_TRANSPARENT, true);
            }

            for (auto& node : nodes)
            {
                node.second->set_material_override(material);
            }
        }
    }
    else if (drawStyle == armarx::viz::data::ModelDrawStyle::ORIGINAL)
    {
        for (auto& node : nodes)
        {
            node.second->set_material_override(nullptr);
        }
    }
}

avi::scene::RobotElementNode::RobotCache&
avi::scene::RobotElementNode::getCache()
{
    static RobotCache cache;
    return cache;
}

std::optional<VirtualRobot::RobotPtr>
avi::scene::RobotElementNode::RobotCache::get(const std::string& project,
                                              const std::string& filename)
{
    Key searched = std::make_pair(project, filename);
    size_t idx = 0;

    for (idx = 0; idx < entries.size(); idx++)
    {
        Entry& entry = entries[idx];

        if (entry.key == searched)
        {
            break;
        }
    }

    if (idx < entries.size())
    {
        Entry entry = std::move(entries[idx]);

        entries.emplace_back(std::move(entry));
        entries.erase(entries.begin() + idx);

        return entries.back().value;
    }

    return {};
}

void
avi::scene::RobotElementNode::RobotCache::set(const std::string& project,
                                              const std::string& filename,
                                              VirtualRobot::RobotPtr robot)
{
    if (entries.size() >= 8)
    {
        entries.erase(entries.begin());
    }

    entries.push_back({{project, filename}, std::move(robot)});
}
