#include "TechnicalControlScheme.h"

#include <cmath>

#include <GlobalConstants.hpp>
#include <Input.hpp>
#include <InputEvent.hpp>
#include <InputEventMouseButton.hpp>
#include <InputEventMouseMotion.hpp>

void
avi::scene::TechnicalControlScheme::_register_methods()
{
    register_method("_ready", &TechnicalControlScheme::_ready);
    register_method("_unhandled_input", &TechnicalControlScheme::_unhandled_input);
    register_method("_process", &TechnicalControlScheme::_process);
}

void
avi::scene::TechnicalControlScheme::_init()
{
    mouseDelta = {0, 0};
    wheelDelta = 0;
    smoothWheelDelta = 0;

    zoomValue = 0.0f;
    zoomDistance = 0.0f;

    viewCamera = nullptr;
}

void
avi::scene::TechnicalControlScheme::setup(avi::core::SettingsProvider& settings)
{
    this->settings = &settings;
}

void
avi::scene::TechnicalControlScheme::_ready()
{
    anchor = get_node<godot::Spatial>("Anchor");
    zoomer = get_node<godot::Spatial>("Anchor/Zoom");

    attachment = get_node<godot::Spatial>("Anchor/Zoom/Attachment");
}

void
avi::scene::TechnicalControlScheme::_unhandled_input(godot::InputEvent* event)
{
    if (!isActive()) return;

    auto* mouseMotionEvent = cast_to<godot::InputEventMouseMotion>(event);
    if (mouseMotionEvent)
    {
        mouseDelta = mouseMotionEvent->get_relative();
    }

    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event);
    if (mouseButtonEvent)
    {
        float change = 0.0f;
        auto idx = mouseButtonEvent->get_button_index();

        if (idx == godot::GlobalConstants::BUTTON_WHEEL_UP)
        {
            change += 1.0f;
        }

        if (idx == godot::GlobalConstants::BUTTON_WHEEL_DOWN)
        {
            change -= 1.0f;
        }

        wheelDelta += change;
    }
}

void
avi::scene::TechnicalControlScheme::_process(float delta)
{
    if (!isActive()) return;

    godot::Vector2 oldMouseDelta = mouseDelta;
    mouseDelta = {0, 0};

    float oldWheelDelta = wheelDelta;
    wheelDelta = 0;

    smoothWheelDelta = godot::Math::lerp(smoothWheelDelta, oldWheelDelta, 0.4f);

    if (isTraveling)
    {
        doTravel(delta);
        return;
    }

    godot::Input* input = godot::Input::get_singleton();

    // Whether the camera should be moved in any way
    bool moveCamera = input->is_action_pressed("avi_camera_mouse") || viewCamera->isMouseLocked();
    // Whether the camera should use alternative movement (e.g. move anchor instead of around anchor)
    bool alternative = input->is_action_pressed("avi_camera_alt");

    if (moveCamera)
    {
        if (!alternative)
        {
            orbit(oldMouseDelta, delta);
        }
        else
        {
            move(oldMouseDelta, delta);
        }
    }

    if (!alternative)
    {
        zoom(smoothWheelDelta, delta);
    }
    else
    {
        roll(smoothWheelDelta, delta);
    }

    anchor->set_transform(anchor->get_transform().orthonormalized());
    zoomer->set_transform(zoomer->get_transform().orthonormalized());
}

void
avi::scene::TechnicalControlScheme::takeOver(avi::scene::ViewCamera* viewCamera)
{
    this->viewCamera = viewCamera;

    godot::Transform target = viewCamera->get_transform();
    viewCamera->set_transform(godot::Transform());

    godot::Transform change = target * zoomer->get_transform().inverse();
    anchor->set_transform(change);

    attachment->add_child(viewCamera);
}

void
avi::scene::TechnicalControlScheme::release()
{
    attachment->remove_child(viewCamera);
    viewCamera->set_transform(anchor->get_transform() * zoomer->get_transform());
    this->viewCamera = nullptr;
}

void
avi::scene::TechnicalControlScheme::reset()
{
    startTravel(godot::Transform(), 0.0f, 0.5f);
}

void
avi::scene::TechnicalControlScheme::focus(godot::AABB box)
{
    godot::Transform transform = anchor->get_transform();
    transform.origin = box.position + box.size * 0.5f;

    float distance = box.size.length() * 3;
    float value = -std::log(distance);

    startTravel(transform, value, 0.5f);
}

void
avi::scene::TechnicalControlScheme::focus(godot::Vector3 point, avi::core::Element* target)
{
    godot::Transform transform = anchor->get_transform();
    transform.origin = point;

    float distance = target->getVolume().size.length() * 3;
    float value = -std::log(distance);

    startTravel(transform, value, 0.5f);
}

godot::Basis
avi::scene::TechnicalControlScheme::getRotationBasis()
{
    return (anchor->get_transform() * zoomer->get_transform().inverse()).basis.inverse();
}

bool
avi::scene::TechnicalControlScheme::isActive()
{
    return viewCamera != nullptr;
}

void
avi::scene::TechnicalControlScheme::orbit(godot::Vector2 rotation, float deltaTime)
{
    const float sensitivity = static_cast<float>(settings->getOrbitSensitivity() * 0.001);
    rotation *= sensitivity * deltaTime;

    anchor->rotate_object_local({0, 1, 0}, -rotation.x);
    anchor->rotate_object_local({1, 0, 0}, -rotation.y);
}

void
avi::scene::TechnicalControlScheme::move(godot::Vector2 movement, float deltaTime)
{
    const float sensitivity = static_cast<float>(settings->getMoveSpeed());
    movement *= sensitivity * deltaTime;
    movement *= zoomDistance;

    godot::Transform transform = anchor->get_transform();
    godot::Basis basis = transform.basis;

    transform.origin += basis.x * -movement.x;
    transform.origin += basis.y * movement.y;

    anchor->set_transform(transform);
}

void
avi::scene::TechnicalControlScheme::zoom(float delta, float deltaTime)
{
    const float sensitivity = static_cast<float>(settings->getZoomSensitivity() * 0.01);
    delta *= sensitivity * deltaTime;

    zoomValue += delta;
    zoomValue = getClampedZoom(zoomValue);
    zoomDistance = std::exp(-1 * zoomValue);

    godot::Transform transform = zoomer->get_transform();
    transform.origin = godot::Vector3(0, 0, zoomDistance);
    zoomer->set_transform(transform);

    float clipPlaneValue;

    if (zoomValue > 4)
    {
        clipPlaneValue = std::pow(4, -zoomValue);
    }
    else
    {
        clipPlaneValue = std::pow(2, -(zoomValue + 4));
    }
    this->viewCamera->setClipPlanes(clipPlaneValue);
}

void
avi::scene::TechnicalControlScheme::roll(float delta, float deltaTime)
{
    const float sensitivity = static_cast<float>(settings->getRollSensitivity() * 0.02);
    delta *= sensitivity * deltaTime;

    anchor->rotate_object_local({0, 0, 1}, delta);
}

void
avi::scene::TechnicalControlScheme::startTravel(godot::Transform anchor, float zoom, float duration)
{
    travel = {.startTransform = this->anchor->get_transform(),
              .endTransform = anchor,
              .startZoom = this->zoomValue,
              .endZoom = getClampedZoom(zoom),
              .duration = duration,
              .remainingTime = duration};

    isTraveling = true;
}

void
avi::scene::TechnicalControlScheme::doTravel(float deltaTime)
{
    travel.remainingTime -= deltaTime;
    float progress = 1.0f - (travel.remainingTime / travel.duration);

    if (travel.remainingTime < 0)
    {
        isTraveling = false;
        progress = 1.0f;
    }

    godot::Transform transform =
        travel.startTransform.interpolate_with(travel.endTransform, progress);
    anchor->set_transform(transform);

    zoomValue = godot::Math::lerp(travel.startZoom, travel.endZoom, progress);
    zoom(0, deltaTime);
}

float
avi::scene::TechnicalControlScheme::getClampedZoom(float zoom)
{
    return std::clamp(zoom, -250.0f, 10.0f);
}
